#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import os
import time
import torch
import logging
import numpy as np
import pandas as pd
from torch import nn
from collections import defaultdict
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score
from .utils import set_device, tensor2flatten_arr
from logai.config_interfaces import Config
from logai.utils.file_utils import read_file
from attr import dataclass
from logai.algorithms.vectorization_algo.forecast_nn import ForecastNNVectorizedDataset
from torch.utils.data import DataLoader


@dataclass
class ForecastBasedNNParams(Config):
    """
    Config for neural representation learning for logs using forecasting based self-supervised tasks.

    :param model_name: name of the model.
    :param metadata_filepath: path to file containing meta data (pretrained token embeddings in case if
        semantic log representations are used in feature type).
    :param output_dir: path to output directory where the model would be dumped.
    :param feature_type: (should be "semantics" or "sequential")type of log feature representations used
        for the log-lines or log-sequences.
    :param label_type: type of label (should be "anomaly" or "next_log") based on whether supervised
        or unsupervised (forcast based) model is being used.
    :param eval_type: (should be "session" or None) whether to aggregate and report the evaluation
        metrics at the level of sessions (based on the span_id in the log data) or at the level of each logline.
    :param topk: the prediction at top-k to consider, when deciding whether an evaluation instance is an anomaly or not.
    :param embedding_dim: dimension of the embedding space. Both for sequential and semantic type feature representation,
        the input log feature representation is passed through an embedding layer which projects it to the embedding_dim.
    :param hidden_size: dimension of the hidden representations.
    :param freeze: whether to freeze the embedding layer to use the pretrained embeddings or to further train it on the given task.
    :param gpu: device number if gpu is used (otherwise -1 or None will use cpu).
    :param patience: number of eval_steps, the model waits for performance on validation data to improve, before early stopping the training.
    :param num_train_epochs: number of training epochs.
    :param batch_size: batch size.
    :param learning_rate: learning rate.
    """

    model_name: str = None
    metadata_filepath: str = None
    output_dir: str = None
    feature_type: str = ""  # sequential, semantics
    label_type: str = ""  # anomaly, next_log
    eval_type: str = "session"  # session, None
    topk: int = 10
    embedding_dim: int = 100
    hidden_size: int = 100
    freeze: bool = False
    gpu: int = None
    patience: int = 5
    num_train_epochs: int = 100
    batch_size: int = 1024
    learning_rate: int = 1e-4


class Embedder(nn.Module):
    """Learnable embedder for embedding loglines.

    :param vocab_size: vocabulary size.
    :param embedding_dim: embedding dimension.
    :param pretrain_matrix: torch.Tensor object containing the pretrained embedding of the vocabulary tokens.
    :param freeze: Freeze embeddings to pretrained ones if set to True, otherwise makes the embeddings learnable.
    """

    def __init__(
        self,
        vocab_size: int,
        embedding_dim: int,
        pretrain_matrix: np.array = None,
        freeze: bool = False,
    ):
        super(Embedder, self).__init__()
        if pretrain_matrix is not None:
            self.embedding_layer = nn.Embedding.from_pretrained(
                pretrain_matrix, padding_idx=1, freeze=freeze
            )
        else:
            self.embedding_layer = nn.Embedding(
                vocab_size, embedding_dim, padding_idx=1
            )

    def forward(self, x):
        return self.embedding_layer(x.long())


class ForecastBasedNN(nn.Module):
    """
    Model for learning log representations through a forecasting based self-supervised task.

    :param config: ForecastBasedNNParams config class for parameters of forecasting based neural log representation models.
    """

    def __init__(self, config: ForecastBasedNNParams):

        super(ForecastBasedNN, self).__init__()
        self.config = config
        self.device = set_device(self.config.gpu)
        self.topk = self.config.topk
        self.feature_type = self.config.feature_type
        self.label_type = self.config.label_type
        self.eval_type = self.config.eval_type
        self.patience = self.config.patience
        self.time_tracker = {}
        self.num_train_epochs = self.config.num_train_epochs
        self.learning_rate = self.config.learning_rate

        self.meta_data = read_file(self.config.metadata_filepath)
        model_save_dirpath = os.path.join(
            self.config.output_dir, "model_" + self.config.model_name
        )
        os.makedirs(model_save_dirpath, exist_ok=True)
        self.model_save_file = os.path.join(model_save_dirpath, "model.ckpt")
        if self.feature_type in ["sequential", "semantics"]:
            self.embedder = Embedder(
                self.meta_data["vocab_size"],
                embedding_dim=self.config.embedding_dim,
                pretrain_matrix=self.meta_data.get("pretrain_matrix", None),
                freeze=self.config.freeze,
            )
        else:
            logging.info(
                f"Unrecognized feature type, except sequentials or semantics, got {self.feature_type}"
            )

    def predict(self, test_loader: DataLoader, dtype: str = "test"):
        """
        Predict method on test data.

        :param test_loader: dataloader (torch.utils.data.DataLoader) for test (or development) dataset.
        :param dtype: can be of type "test" or "dev" based on which the predict method is called for.
        :return: dict object containing the overall evaluation metrics for test (or dev) data.
        """
        logging.info("Evaluating {} data.".format(dtype))

        if self.label_type == "next_log":
            return self.__predict_next_log(test_loader, dtype=dtype)
        elif self.label_type == "anomaly":
            return self.__predict_anomaly(test_loader, dtype=dtype)

    def __predict_anomaly(self, test_loader: DataLoader, dtype: str = "test"):

        model = self.eval()  # set to evaluation mode
        with torch.no_grad():
            y_pred = []
            store_dict = defaultdict(list)
            infer_start = time.time()
            if dtype == "dev":
                epoch_loss = 0
                batch_cnt = 0
                for batch_input in test_loader:
                    batch_input = self.__input2device(batch_input)
                    return_dict = model.forward(batch_input)
                    epoch_loss += return_dict["loss"].item()
                    batch_cnt += 1
                epoch_loss = epoch_loss / batch_cnt
                logging.info("Dev Loss: {} ".format(epoch_loss))
                return {"loss": epoch_loss}
            elif dtype == "test":
                for batch_input in test_loader:
                    return_dict = model.forward(self.__input2device(batch_input))
                    y_prob, y_pred = return_dict["y_pred"].max(dim=1)
                    store_dict[ForecastNNVectorizedDataset.session_idx].extend(
                        tensor2flatten_arr(
                            batch_input[ForecastNNVectorizedDataset.session_idx]
                        )
                    )
                    store_dict[ForecastNNVectorizedDataset.window_anomalies].extend(
                        tensor2flatten_arr(
                            batch_input[ForecastNNVectorizedDataset.window_anomalies]
                        )
                    )
                    store_dict["window_preds"].extend(tensor2flatten_arr(y_pred))
                infer_end = time.time()
                logging.info(
                    "Finish inference. [{:.2f}s]".format(infer_end - infer_start)
                )
                self.time_tracker["test"] = infer_end - infer_start

                store_df = pd.DataFrame(store_dict)
                use_cols = [
                    ForecastNNVectorizedDataset.session_idx,
                    ForecastNNVectorizedDataset.window_anomalies,
                    "window_preds",
                ]
                session_df = (
                    store_df[use_cols]
                    .groupby(ForecastNNVectorizedDataset.session_idx, as_index=False)
                    .sum()
                )
                pred = (session_df[f"window_preds"] > 0).astype(int)
                y = (
                    session_df[ForecastNNVectorizedDataset.window_anomalies] > 0
                ).astype(int)

                assert len(store_dict["window_preds"]) == len(
                    test_loader.dataset
                ), "Length of predictions {} does not match length of test dataset {}".format(
                    len(store_dict["window_preds"]), len(test_loader.dataset)
                )
                eval_results = {
                    "f1": f1_score(y, pred),
                    "rc": recall_score(y, pred),
                    "pc": precision_score(y, pred),
                    "acc": accuracy_score(y, pred),
                    "pred": pred,
                    "true": y,
                }
                logging.info(
                    "Best result: f1: {} rc: {} pc: {}".format(
                        eval_results["f1"], eval_results["rc"], eval_results["pc"]
                    )
                )
                return eval_results

    def __predict_next_log(self, test_loader: DataLoader, dtype: str = "test"):
        model = self.eval()  # set to evaluation mode
        with torch.no_grad():
            y_pred = []
            store_dict = defaultdict(list)
            infer_start = time.time()

            if dtype == "dev":
                epoch_loss = 0
                batch_cnt = 0
                y_true_prob_list = []
                y_true_list = []
                y_pred_list = []
                for batch_input in test_loader:
                    batch_input = self.__input2device(batch_input)
                    return_dict = model.forward(batch_input)
                    epoch_loss += return_dict["loss"].item()
                    y_pred = return_dict["y_pred"]
                    y_true = batch_input[ForecastNNVectorizedDataset.window_labels]
                    y_true_prob = torch.gather(y_pred, 1, y_true.unsqueeze(dim=-1))
                    y_pred = torch.argmax(y_pred, dim=-1)
                    y_true_prob_list.extend(y_true_prob.data.cpu().numpy())
                    y_true_list.extend(y_true.data.cpu().numpy())
                    y_pred_list.extend(y_pred.data.cpu().numpy())
                    batch_cnt += 1
                epoch_loss = epoch_loss / batch_cnt
                logging.info("Dev Loss: {}".format(epoch_loss))
                y_true_prob_list = np.array(y_true_prob_list)
                num_y = len(y_true_list)
                y_true_list = np.array(y_true_list)
                y_pred_list = np.array(y_pred_list)

                y_correct = np.sum(np.equal(y_true_list, y_pred_list, dtype=np.int))
                logging.info(
                    "Dev acc @ top-1: {}  correct: {} out of {}".format(
                        float(y_correct) / num_y, y_correct, num_y
                    )
                )
                return {"loss": epoch_loss}
            elif dtype == "test":
                for batch_input in test_loader:
                    batch_input = self.__input2device(batch_input)
                    return_dict = model.forward(batch_input)
                    y_pred = return_dict["y_pred"]
                    y_prob_topk, y_pred_topk = torch.topk(y_pred, self.topk)  # b x topk
                    store_dict[ForecastNNVectorizedDataset.session_idx].extend(
                        tensor2flatten_arr(
                            batch_input[ForecastNNVectorizedDataset.session_idx]
                        )
                    )
                    store_dict[ForecastNNVectorizedDataset.window_anomalies].extend(
                        tensor2flatten_arr(
                            batch_input[ForecastNNVectorizedDataset.window_anomalies]
                        )
                    )
                    store_dict[ForecastNNVectorizedDataset.window_labels].extend(
                        tensor2flatten_arr(
                            batch_input[ForecastNNVectorizedDataset.window_labels]
                        )
                    )
                    store_dict["x"].extend(
                        batch_input[ForecastNNVectorizedDataset.features]
                        .data.cpu()
                        .numpy()
                    )
                    store_dict["y_pred_topk"].extend(y_pred_topk.data.cpu().numpy())
                    store_dict["y_prob_topk"].extend(y_prob_topk.data.cpu().numpy())
                infer_end = time.time()
                logging.info("Finish inference. [{}s]".format(infer_end - infer_start))

                assert len(store_dict["x"]) == len(
                    test_loader.dataset
                ), "Length of predictions {} does not match length of test dataset {}".format(
                    len(store_dict["x"]), len(test_loader.dataset)
                )
                self.time_tracker["test"] = infer_end - infer_start
                store_df = pd.DataFrame(store_dict)
                best_result = None
                best_f1 = -float("inf")
                count_start = time.time()

                topkdf = pd.DataFrame(store_df["y_pred_topk"].tolist())
                logging.info("Calculating acc sum.")
                not_hit = None
                for col in sorted(topkdf.columns):
                    topk = col + 1
                    if not_hit is not None:
                        not_hit = np.logical_and(
                            not_hit,
                            np.array(
                                (
                                    topkdf[col]
                                    != store_df[
                                        ForecastNNVectorizedDataset.window_labels
                                    ]
                                ).astype(int)
                            ),
                        )
                    else:
                        not_hit = np.array(
                            (
                                topkdf[col]
                                != store_df[ForecastNNVectorizedDataset.window_labels]
                            ).astype(int)
                        )
                    store_df["window_pred_anomaly_{}".format(topk)] = not_hit

                logging.info("Finish generating store_df.")

                if self.eval_type == "session":
                    use_cols = [
                        ForecastNNVectorizedDataset.session_idx,
                        ForecastNNVectorizedDataset.window_anomalies,
                    ] + [
                        f"window_pred_anomaly_{topk}"
                        for topk in range(1, self.topk + 1)
                    ]
                    session_df = (
                        store_df[use_cols]
                        .groupby(
                            ForecastNNVectorizedDataset.session_idx, as_index=False
                        )
                        .sum()
                    )
                else:
                    session_df = store_df

                for topk in range(1, self.topk + 1):
                    pred = (session_df[f"window_pred_anomaly_{topk}"] > 0).astype(int)
                    y = (
                        session_df[ForecastNNVectorizedDataset.window_anomalies] > 0
                    ).astype(int)
                    eval_results = {
                        "f1": f1_score(y, pred),
                        "rc": recall_score(y, pred),
                        "pc": precision_score(y, pred),
                        "pred": pred,
                        "true": y,
                    }
                    if eval_results["f1"] >= best_f1:
                        best_result = eval_results
                        best_f1 = eval_results["f1"]
                count_end = time.time()
                logging.info("Finish counting [{}s]".format(count_end - count_start))
                logging.info(
                    "Best result: f1: {} rc: {} pc: {}".format(
                        best_result["f1"], best_result["rc"], best_result["pc"]
                    )
                )
                return best_result

    def __input2device(self, batch_input: dict):
        return {k: v.to(self.device) for k, v in batch_input.items()}

    def save_model(self):
        """Saving model to file as specified in config"""
        logging.info("Saving model to {}".format(self.model_save_file))
        try:
            torch.save(
                self.state_dict(),
                self.model_save_file,
                _use_new_zipfile_serialization=False,
            )
        except Exception as e:
            torch.save(self.state_dict(), self.model_save_file)

    def load_model(self, model_save_file: str = ""):
        """Loading model from file.

        :param model_save_file: path to file where model would be saved.
        """
        logging.info("Loading model from {}".format(self.model_save_file))
        self.load_state_dict(torch.load(model_save_file, map_location=self.device))

    def fit(self, train_loader: DataLoader, dev_loader: DataLoader = None):
        """
        Fit method for training model

        :param train_loader: dataloader (torch.utils.data.DataLoader) for the train dataset.
        :param dev_loader: dataloader (torch.utils.data.DataLoader) for the train dataset. Defaults to None, for which no evaluation is run.
        :return: dict containing the best loss on dev dataset.
        """
        self.to(self.device)
        logging.info(
            "Start training on {} batches with {}.".format(
                len(train_loader), self.device
            )
        )
        best_loss = float("inf")
        best_results = None
        worse_count = 0
        for epoch in range(1, self.num_train_epochs + 1):
            epoch_time_start = time.time()
            model = self.train()
            optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)

            batch_cnt = 0
            epoch_loss = 0
            for batch_input in train_loader:
                loss = model.forward(self.__input2device(batch_input))["loss"]
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                epoch_loss += loss.item()
                batch_cnt += 1
                if batch_cnt % 100 == 0:
                    logging.info(
                        "Batch {}, training loss : {}".format(
                            batch_cnt, epoch_loss / batch_cnt
                        )
                    )
            epoch_loss = epoch_loss / batch_cnt
            epoch_time_elapsed = time.time() - epoch_time_start
            logging.info(
                "Epoch {}/{}, training loss: {} [{}s]".format(
                    epoch, self.num_train_epochs, epoch_loss, epoch_time_elapsed
                )
            )
            self.time_tracker["train"] = epoch_time_elapsed

            if dev_loader is not None and (epoch % 1 == 0):
                eval_results = self.predict(dev_loader, dtype="dev")
                if eval_results["loss"] < best_loss:
                    best_loss = eval_results["loss"]
                    best_results = eval_results
                    best_results["converge"] = int(epoch)
                    self.save_model()
                    worse_count = 0
                else:
                    worse_count += 1
                    if worse_count >= self.patience:
                        logging.info("Early stop at epoch: {}".format(epoch))
                        break
        self.load_model(self.model_save_file)
        return best_results



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from .partitioner import Partitioner, PartitionerConfig
from logai.information_extraction.feature_extractor import (
    FeatureExtractor,
    FeatureExtractorConfig,
)
import pandas as pd
from logai.utils import constants
from logai.config_interfaces import Config

from attr import dataclass


@dataclass
class OpenSetPartitionerConfig(Config):
    """Config for Partitioner for open log datasets.

    :param sliding_window: The size of sliding window.
    :param session_window: A boolean flag whether to use session based partitioning or not.
    :param logsequence_delim: The delimiter string for concatenating log sequences.
    """

    sliding_window: int = 0
    session_window: bool = True
    logsequence_delim: str = "[SEP]"


class OpenSetPartitioner:
    """Partitioner class for Open log datasets.
        
    :param config: A config object specifying parameters of log partititoning for open log datasets.
    """
    def __init__(self, config: OpenSetPartitionerConfig):
        
        self.config = config

        if config.sliding_window > 0:
            partitioner_config = PartitionerConfig.from_dict(
                {
                    "sliding_window": config.sliding_window,
                    "sep_token": config.logsequence_delim,
                    "group_by_category": [constants.SPAN_ID],
                    "exclude_last_window": True,
                    "exclude_smaller_windows": True,
                }
            )
            self.partitioner = Partitioner(partitioner_config)
        elif config.session_window:
            fe_config = FeatureExtractorConfig.from_dict(
                {
                    "sliding_window": 20,
                    "steps": 20,
                    "group_by_category": [constants.SPAN_ID],
                }
            )
            self.feature_extractor = FeatureExtractor(fe_config)

    def _get_group_sliding_window(
        self, logrecord_df, logline_col_name=constants.LOGLINE_NAME
    ):
        logrecord_df[logline_col_name] = logrecord_df[logline_col_name].astype(str)
        return self.partitioner.group_sliding_window(logrecord_df, logline_col_name)

    def _get_sliding_window_label(self, log_data):
        partitioned_data = self._get_group_sliding_window(
            log_data, logline_col_name=constants.LABELS
        )
        partitioned_labels = partitioned_data[constants.LABELS].apply(
            lambda x: int("1" in x)
        )
        partitioned_labels_nextdata = self._get_next_data_succeeding_sliding_window(
            log_data, partitioned_data, constants.LABELS
        ).astype(int)
        partitioned_labels = (
            (partitioned_labels_nextdata + partitioned_labels) > 0
        ).astype(int)
        return partitioned_labels

    def _get_next_data_succeeding_sliding_window(self, data, sliding_windows, field):
        data_groupbyid = {
            k: list(v) for k, v in data.groupby(by=[constants.SPAN_ID])[field]
        }
        nextdata = []
        for id, groupdata in sliding_windows.groupby(by=[constants.SPAN_ID])[field]:
            nextdata.extend(data_groupbyid[id][-len(groupdata) :])
        nextdata = pd.Series(nextdata).astype(data[field].dtype)
        return nextdata

    def generate_sliding_window(self, logrecord):
        """Method to generate sliding window based log sequences from a logrecord object.
        
        :param logrecord: A log record object to be partitioned into sliding windows.
        :return: LogRecordObject where the body of logrecord object contains the generated log sequences.
        """
        log_data = logrecord.to_dataframe()
        partitioned_data = self._get_group_sliding_window(
            log_data, logline_col_name=constants.LOGLINE_NAME
        )
        partitioned_nextdata = self._get_next_data_succeeding_sliding_window(
            log_data, partitioned_data, constants.LOGLINE_NAME
        )
        partitioned_labels = self._get_sliding_window_label(log_data)
        partitioned_ids = partitioned_data[constants.SPAN_ID]
        partitioned_loglines = partitioned_data[constants.LOGLINE_NAME]
        logrecord.body = pd.DataFrame({constants.LOGLINE_NAME: partitioned_loglines})
        logrecord.attributes = pd.DataFrame(
            {constants.NEXT_LOGLINE_NAME: partitioned_nextdata}
        )
        logrecord.labels = pd.DataFrame({constants.LABELS: partitioned_labels})
        logrecord.span_id = pd.DataFrame({constants.SPAN_ID: partitioned_ids})
        return logrecord

    def generate_session_window(self, logrecord):
        """Method to generate session window based log sequences from a logrecord object given some.
        ids at the logline level
        
        :param logrecord: A log record object to be partitioned into session windows.
        :return: LogRecordObject where the body of logrecord object contains the generated log sequences.
        """

        partitioned_data = self.feature_extractor.convert_to_counter_vector(
            log_pattern=logrecord.body[constants.LOGLINE_NAME],
            attributes=logrecord.span_id.join(logrecord.labels),
            timestamps=logrecord.timestamp[constants.LOG_TIMESTAMPS],
        )
        partitioned_loglines = partitioned_data[constants.LOGLINE_NAME].apply(
            lambda x: self.config.logsequence_delim.join(x)
        )
        partitioned_labels = partitioned_data[constants.LABELS].apply(
            lambda x: int(sum(x) > 0)
        )
        partitioned_ids = partitioned_data[constants.SPAN_ID]
        logrecord.body = pd.DataFrame({constants.LOGLINE_NAME: partitioned_loglines})
        logrecord.labels = pd.DataFrame({constants.LABELS: partitioned_labels})
        logrecord.span_id = pd.DataFrame({constants.SPAN_ID: partitioned_ids})
        return logrecord

    def partition(self, logrecord):
        """Wrapper function for applying partitioning on a logrecord object based on the Config parameters.
        
        :param logrecord: A log record object to be partitioned into session or sliding windows.
        :return: LogRecordObject where the body of logrecord object contains the generated log sequences.
        """
        if self.config.sliding_window > 0:
            logrecord = self.generate_sliding_window(logrecord)
        elif self.config.session_window:
            logrecord = self.generate_session_window(logrecord)
        return logrecord



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
"""
Module that includes common data manipulation functions to be applied by pandas dataframes.
"""

import numpy as np
import pandas as pd
from merlion.utils import TimeSeries

from logai.utils import constants


def pad(x, max_len: np.array, padding_value: int = 0):
    """Method to trim or pad any 1-d numpy array to a given max length with the given padding value
    
    :param x: (np.array): given 1-d numpy array to be padded/trimmed
    :param max_len: (int): maximum length of padded/trimmed output
    :param padding_value: (int, optional): padding value. Defaults to 0.
    :return: np.array: padded/trimmed numpy array
    """
    flattened_vector = x
    fill_size = max_len - len(flattened_vector)
    if fill_size > 0:
        fill_zeros = np.full(fill_size, fill_value=padding_value)
        return np.concatenate((flattened_vector, fill_zeros), axis=0)
    else:
        return flattened_vector[:max_len]


def get_parameter_list(row):
    parameter_list = []
    if not isinstance(row[constants.LOGLINE_NAME], str) or not isinstance(
        row[constants.PARSED_LOGLINE_NAME], str
    ):
        return parameter_list
    ll = row[constants.LOGLINE_NAME].split()
    pp = row[constants.PARSED_LOGLINE_NAME].split()
    buffer = []

    i = 0
    j = 0
    consec_pattern = False
    while i < len(ll) and j < len(pp):
        # print(ll[i], pp[j])
        if ll[i] == pp[j]:
            if buffer:
                parameter_list.append(" ".join(buffer))
                buffer = []
            consec_pattern = False
            i += 1
            j += 1
        elif pp[j] == "*":
            if consec_pattern:
                parameter_list.append(" ".join(buffer))
                buffer = [ll[i]]
            else:
                buffer.append(ll[i])
            consec_pattern = True
            i += 1
            j += 1
        else:
            buffer.append(ll[i])
            i += 1
    if buffer:
        if i < len(ll):
            parameter_list.append(" ".join(buffer + ll[i:]))
        else:
            parameter_list.append(" ".join(buffer))
    return parameter_list


def pd_to_timeseries(log_features: pd.Series):
    """
    Convert pandas.DataFrame to merlion.TimeSeries for log counter vectors.

    :param log_features: log feature dataframe must only contain two columns
      ['timestamp': datetime, constants.LOGLINE_COUNTS: int].
    :return: merlion.TimeSeries type.
    """
    ts_df = log_features[constants.LOG_COUNTS]
    ts_df.index = log_features[constants.LOG_TIMESTAMPS]
    time_series = TimeSeries.from_pd(ts_df)
    return time_series



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from dataclasses import dataclass, field
import pathlib
import json
from logai.utils import constants
import logging


@dataclass
class LogRecordObject:
    """
    Log record object data model, compatible with log and event record definition in OpenTelemetry
    https://opentelemetry.io/docs/reference/specification/logs/data-model/#log-and-event-record-definition.
    
    :param timestamp: The timestamp information of the log data.
    :param attributes: The attributes of the log data (typically structured data with quantitative or categorical fields).
    :param resource: The field denoting data source information generating the log data.
    :param trace_id: The request trace id associated with the log data, if any.
    :param span_id: The request span id associated with the log data, if any.
    :param severity_text: The severity description or log level information.
    :param severity_number: The severity number indicating log level.
    :param body: The body of the log record, which contains the main information of the log. It can be consisting
        of either unstructured, semi-structured or structured information.
    :param labels: Any associated label information with the log (for e.g. binary anomaly label indicating
        whether each line is anomalous or not).
    :param _index: The indices of the log data.
    """

    timestamp: pd.DataFrame = pd.DataFrame()
    attributes: pd.DataFrame = pd.DataFrame()
    resource: pd.DataFrame = pd.DataFrame()
    trace_id: pd.DataFrame = pd.DataFrame()
    span_id: pd.DataFrame = pd.DataFrame()
    severity_text: pd.DataFrame = pd.DataFrame()
    severity_number: pd.DataFrame = pd.DataFrame()
    body: pd.DataFrame = pd.DataFrame()
    labels: pd.DataFrame = pd.DataFrame()
    _index: np.array = field(init=False)

    def __post_init__(self):
        self._index = pd.DataFrame(self.body.index.values)

        for field in self.__dataclass_fields__:
            field_content = getattr(self, field)
            if not field_content.empty:
                if not field_content.index.equals(self._index.index):
                    raise IndexError(
                        "Index of {} should match Index of this object".format(field)
                    )

    def _meta_data(self):
        meta_data_dict = {}
        for field in self.__dataclass_fields__:
            field_content = getattr(self, field)
            if field == "_index":
                continue
            if type(field_content) == pd.DataFrame and not field_content.empty:
                columns = list(field_content.keys())
                meta_data_dict[field] = columns
        return meta_data_dict

    def to_dataframe(self):
        """
        Generates pandas.DataFrame from LogRecordType.
        """
        if self.body.empty:
            return None
        df = self.body
        for field in self.__dataclass_fields__:
            if field == "body":
                continue
            field_content = getattr(self, field)
            if not field_content.empty:
                df = df.join(field_content)

        return df

    @classmethod
    def from_dataframe(cls, data: pd.DataFrame, meta_data: dict = None):
        """
        Converts pandas.DataFrame to log record object.
        
        :param data: The log data in pandas dataframe.
        :param meta_data: A dictionary that maps data.columns to fields of LogRecordObject.
        :return: A LogRecordObject object.
        """
        if meta_data is None and not data.empty:
            logbody = data.apply(" ".join, axis=1)
            return LogRecordObject(body=logbody)

        logrecord = LogRecordObject()
        for key, value in meta_data.items():
            if key not in logrecord.__dataclass_fields__.keys():
                raise KeyError(
                    "{} is not a field in LogRecordObject. The valid fields: {}".format(
                        key, logrecord.__dataclass_fields__.keys()
                    )
                )
            logrecord.__setattr__(key, data[value])
        logrecord.__post_init__()
        return logrecord

    def save_to_csv(self, filepath: str):
        """
        Saves a log record object to file.
        
        :param filepath: The absolute path to filename where the logrecord object would be saved.
        """
        f = pathlib.Path(filepath)
        filepath_metadata = filepath.replace(f.suffix, "_metadata.json")
        if f.suffix == ".csv":
            self.to_dataframe().to_csv(filepath, encoding="utf8")
        else:
            raise Exception("Only supports csv format")
        with open(filepath_metadata, "w") as fp:
            json.dump(self._meta_data(), fp)

    @classmethod
    def load_from_csv(cls, filepath):
        f = pathlib.Path(filepath)
        filepath_metadata = filepath.replace(f.suffix, "_metadata.json")
        data = pd.read_csv(filepath)
        meta_data = json.load(open(filepath_metadata))
        return cls.from_dataframe(data=data, meta_data=meta_data)

    def select_by_index(self, indices: list, inplace: bool = False):
        """Selects a subset of a logrecord object based on selected indices.
        
        :param indices: A list of indices to select inplace (bool, optional): performs operation inplace or not.
        :return: LogRecordObject: The resulting logr ecord object created from the selected indices.
        """
        if not inplace:
            target = LogRecordObject()
        else:
            target = self
        for key in self.__dataclass_fields__.keys():
            val = getattr(self, key)
            if type(val) == pd.DataFrame and not val.empty:
                val = val[val.index.isin(indices)]
            target.__setattr__(key, val)
        return target

    def filter_by_index(self, indices: list, inplace: bool = False):
        """Selects a subset of a logrecord object by removing certain indices.
        
        :param indices: A list of indices to remove inplace (bool, optional): performs operation inplace or not.
        :return: The resulting log record object created after removing the indices.
        """
        if not inplace:
            target = LogRecordObject()
        else:
            target = self
        for key in self.__dataclass_fields__.keys():
            val = getattr(self, key)
            if type(val) == pd.DataFrame and not val.empty:
                val = val[~val.index.isin(indices)]
            target.__setattr__(key, val)
        return target

    def dropna(self):
        """Method to drop entries containing NaN or null values in the logrecord object.
        
        :return: The modified logrecord object after removing entries with NaN or null values.
        """
        null_body = self.body.isnull()
        null_body = null_body[null_body[constants.LOGLINE_NAME] == True]
        null_indices = list(null_body.index)
        if len(null_indices) > 0:
            logging.info(
                "Removed {} indices with null value ".format(len(null_indices))
            )
            return self.filter_by_index(null_indices, inplace=True)
        else:
            return self



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
"""
This is wrapping the logpai/logparser implementation of AEL algorithm
link: https://github.com/logpai/logparser/blob/master/logparser/AEL/AEL.py.
"""

import re
import hashlib
import pandas as pd
from collections import defaultdict
from functools import reduce
from attr import dataclass

from logai.algorithms.algo_interfaces import ParsingAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


class Event:
    """Event class to wrap log events.
    """
    def __init__(self, logidx, Eventstr=""):
        self.id = hashlib.md5(Eventstr.encode("utf-8")).hexdigest()[0:8]
        self.logs = [logidx]
        self.Eventstr = Eventstr
        self.EventToken = Eventstr.split()
        self.merged = False

    def refresh_id(self):
        """Generates id for a log event using the hashing function.
        """
        self.id = hashlib.md5(self.Eventstr.encode("utf-8")).hexdigest()[0:8]


@dataclass
class AELParams(Config):
    """Parameters for the AEL Log Parsing algorithm. For more details see 
    https://github.com/logpai/logparser/tree/master/logparser/AEL.

    :param rex: A rex string.
    :param minEventCount: The minimum event count.
    :param merge_percent: The merge percentage.
    :param keep_para: Whether to keep parameters.
    """
    rex: str = None
    minEventCount: int = 2
    merge_percent: int = 1
    keep_para: bool = True


@factory.register("parsing", "ael", AELParams)
class AEL(ParsingAlgo):
    def __init__(self, params: AELParams):
        self.rex = params.rex
        self.minEventCount = params.minEventCount
        self.merge_percent = params.merge_percent
        self.df_log = None
        self.logname = None
        self.merged_events = []
        self.bins = defaultdict(dict)
        self.keep_para = params.keep_para

    def fit(self, loglines: pd.DataFrame):
        """Fit method to train log parser on given log data.
        Since AEL Log Parser does not require any training, this method is empty.
        """
        pass

    def parse(self, loglines: pd.Series) -> pd.Series:
        """Parse method to run log parser on given log data.

        :param loglines: The raw log data to be parsed.
        :returns: The parsed log data.
        """
        self.logname = "logname"
        self.load_data(loglines)
        self.tokenize()
        self.categorize()
        self.reconcile()

        templateL = [0] * self.df_log.shape[0]

        for event in self.merged_events:
            for logidx in event.logs:
                templateL[logidx] = event.Eventstr

        return pd.Series(templateL, index=loglines.index)

    def tokenize(self):
        """
        Puts logs into bins according to (# of '<*>', # of token).
        """
        for idx, log in self.df_log["Content_"].iteritems():
            para_count = 0

            tokens = log.split()
            for token in tokens:
                if token == "<*>":
                    para_count += 1

            if "Logs" not in self.bins[(len(tokens), para_count)]:
                self.bins[(len(tokens), para_count)]["Logs"] = [idx]
            else:
                self.bins[(len(tokens), para_count)]["Logs"].append(idx)

    def categorize(self):
        """
        Categorizes templates bin by bin.
        """
        for key in self.bins:
            abin = self.bins[key]
            abin["Events"] = []

            for logidx in abin["Logs"]:
                log = self.df_log["Content_"].loc[logidx]
                matched = False
                for event in abin["Events"]:
                    if log == event.Eventstr:
                        matched = True
                        event.logs.append(logidx)
                        break
                if not matched:
                    abin["Events"].append(Event(logidx, log))

    def reconcile(self):
        """
        Merges events if a bin has too many events.
        """
        for key in self.bins:
            abin = self.bins[key]
            if len(abin["Events"]) > self.minEventCount:
                tobeMerged = []
                for e1 in abin["Events"]:
                    if e1.merged:
                        continue
                    e1.merged = True
                    tobeMerged.append([e1])

                    for e2 in abin["Events"]:
                        if e2.merged:
                            continue
                        if self.has_diff(e1.EventToken, e2.EventToken):
                            tobeMerged[-1].append(e2)
                            e2.merged = True
                for Es in tobeMerged:
                    merged_event = reduce(self.merge_event, Es)
                    merged_event.refresh_id()
                    self.merged_events.append(merged_event)
            else:
                for e in abin["Events"]:
                    self.merged_events.append(e)

    def merge_event(self, e1, e2):
        """Method to merge two events.

        :param e1: The first event to merge (merged in-place).
        :param e2: The second event to merge.
        :returns: The merged event.
        """
        for pos in range(len(e1.EventToken)):
            if e1.EventToken[pos] != e2.EventToken[pos]:
                e1.EventToken[pos] = "<*>"

        e1.logs.extend(e2.logs)
        e1.Eventstr = " ".join(e1.EventToken)

        return e1

    def has_diff(self, tokens1: list, tokens2: list):
        """Method to check if there is significant different between two given token sequences.

        :param tokens1: The first token sequence.
        :param tokens2: The second token sequence.
        :returns: 0 if no significant difference between given token sequences else 1.
        """
        diff = 0
        for idx in range(len(tokens1)):
            if tokens1[idx] != tokens2[idx]:
                diff += 1
        return True if 0 < diff * 1.0 / len(tokens1) <= self.merge_percent else False

    def load_data(self, loglines: pd.Series):
        """Method to load log data (pandas Series object) to a format compatible for parsing.

        :param loglines: The log data to be parsed.
        """
        def preprocess(log):
            if self.rex:
                for currentRex in self.rex:
                    log = re.sub(currentRex, "<*>", log)
            return log

        self.df_log = pd.DataFrame(loglines)
        self.df_log["Content_"] = self.df_log[loglines.name].map(preprocess)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
import torch
from typing import Tuple, Dict
from attr import dataclass
from datasets import Dataset, load_metric
from sklearn.model_selection import train_test_split
from torch.optim import AdamW
from torch.utils.data import DataLoader
from transformers import (
    AutoModelForSequenceClassification,
    TrainingArguments,
    Trainer,
    AutoTokenizer,
)

from logai.config_interfaces import Config
from logai.utils import constants


@dataclass
class TransformerAlgoConfig(Config):
    """Config class for Transformer based model for log classification tasks.
    """
    tokenizer_config: dict = {"name": "auto", "model": "bert-base-cased"}
    trainer_config: dict = {}


class LogDataset(torch.utils.data.Dataset):
    """Wrapper class for Log Dataset, to wrap over torch Dataset class.
    """
    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        return len(self.labels)


class TransformerAlgo:
    """HuggingFace Transformer based Pretrained Language model (e.g. "bert-base-cased"), 
    with a sequence classifier head for any supervised log classification task.
    For e.g. log anomaly detection is one type of log classfication task where the labels
    are Normal (Label 0) or Anomalous (Label 1). Currently it supports only binary 
    classification, to change this `num_labels` of AutoModelForSequenceClassification 
    has to be changed accordingly along with the prediction logic in predict method.
    """
    def __init__(self, config: TransformerAlgoConfig):
        self.config = config
        self.model = AutoModelForSequenceClassification.from_pretrained(
            self.config.tokenizer_config["model"], num_labels=2
        )
        self.trainer = None
        self.tokenizer = None
        self.metric = load_metric("accuracy")

        return

    def save(self, output_dir: str):
        """Save model in given directory.

        :param output_dir: The path to output directory where model should be dumped.
        """
        self.trainer.save_model(output_dir)
        return

    def train(self, train_logs: pd.Series, train_labels: pd.Series):
        """Train method for Transformer based pretrained language model with
        a sequence classification head for supervised log classification task. 
        Internally this method also splits the available training logs into train and dev data.

        :param train_logs: The training log vectors data (after LogVectorizer).
        :param train_labels: The training label data.
        """
        train_logs = train_logs.rename(constants.LOG_EVENTS)
        if not self.tokenizer:
            if self.config.tokenizer_config["name"] == "auto":
                self.tokenizer = AutoTokenizer.from_pretrained(
                    self.config.tokenizer_config["model"]
                )
            else:
                raise ValueError("Tokenizer is not supported.")

        train_logs, val_logs, train_labels, val_labels = train_test_split(
            train_logs, train_labels, test_size=0.2
        )

        train_df = pd.concat((train_logs, train_labels), axis=1)
        val_df = pd.concat((val_logs, val_labels), axis=1)

        train_dataset = Dataset.from_pandas(train_df)
        val_dataset = Dataset.from_pandas(val_df)

        tokenized_train_datasets = train_dataset.map(
            self._tokenize_auto_tokenizer, batched=True
        )
        tokenized_val_datasets = val_dataset.map(
            self._tokenize_auto_tokenizer, batched=True
        )

        training_args = TrainingArguments(
            output_dir="test_trainer",
            evaluation_strategy="epoch",
            num_train_epochs=2,
            learning_rate=1e-4,
            max_steps=10,
            do_eval=False,
            do_predict=False,
            per_device_train_batch_size=5,
        )

        self.trainer = Trainer(
            model=self.model,
            args=training_args,
            train_dataset=tokenized_train_datasets,
            eval_dataset=tokenized_val_datasets,
            # compute_metrics=self._compute_metrics,
        )

        self.trainer.train()

    def train_with_native_torch(self, train_logs: pd.Series, train_labels: pd.Series):
        """
        Train models in native torch way.

        :param train_logs: The training log features data (after LogVectorizer).
        :param train_labels: The label data for training logs.
        """
        if not self.tokenizer:
            if self.config.tokenizer_config["name"] == "auto":
                self.tokenizer = AutoTokenizer.from_pretrained(
                    self.config.tokenizer_config["model"]
                )
            else:
                raise ValueError("Tokenizer is not supported.")

        train_logs, val_logs, train_labels, val_labels = train_test_split(
            train_logs, train_labels, test_size=0.2
        )

        train_encodings = self.tokenizer(
            train_logs.tolist(), truncation=True, padding=True
        )

        train_dataset = LogDataset(train_encodings, train_labels.tolist())

        self._train(train_dataset=train_dataset)
        self.model.eval()
        return

    def predict(self, test_logs: pd.Series, test_labels: pd.Series) -> Tuple[pd.Series, np.ndarray, Dict[str, float]]:
        """Predict method for running evaluation on test log data.

        :param test_logs: The test log features data (output of LogVectorizer).
        :param test_labels: The labels of test log data.
        :return: - res (pd.Series): Predicted test labels as pandas Series object.
            - label_ids (`np.ndarray`, *optional*): True test labels (if the dataset contained some).
            - metrics (`Dict[str, float]`, *optional*): The potential dictionary of metrics.

        """
        test_logs = test_logs.rename(constants.LOG_EVENTS)
        if not self.tokenizer:
            if self.config.tokenizer_config["name"] == "auto":
                self.tokenizer = AutoTokenizer.from_pretrained(
                    self.config.tokenizer_config["model"]
                )
            else:
                raise ValueError("Tokenizer is not supported.")

        if not test_labels.empty:
            test_df = pd.concat((test_logs, test_labels), axis=1)
        else:
            test_df = pd.DataFrame(test_logs)
        test_dataset = Dataset.from_pandas(test_df)

        tokenized_test_datasets = test_dataset.map(
            self._tokenize_auto_tokenizer, batched=True
        )

        if self.trainer:
            predictions, label_ids, metrics = self.trainer.predict(
                tokenized_test_datasets
            )

            res = pd.Series(
                [0 if p[0] > p[1] else 1 for p in predictions], test_logs.index
            )
            return res, label_ids, metrics
        else:
            raise RuntimeError("No trainer found.")

    def _tokenize_auto_tokenizer(self, examples):
        return self.tokenizer(
            examples[constants.LOG_EVENTS], padding="max_length", truncation=True
        )

    def _compute_metrics(self, eval_pred):
        logits, labels = eval_pred
        predictions = np.argmax(logits, axis=-1)
        return self.metric.compute(predictions=predictions, references=labels)

    def _train(self, train_dataset):
        device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )

        model = AutoModelForSequenceClassification.from_pretrained(
            self.config.tokenizer_config["model"]
        )
        model.to(device)
        model.train()

        train_loader = DataLoader(train_dataset, batch_size=16, shuffle=True)

        optim = AdamW(model.parameters(), lr=5e-5)

        for epoch in range(3):
            for batch in train_loader:
                optim.zero_grad()
                input_ids = batch["input_ids"].to(device)
                attention_mask = batch["attention_mask"].to(device)
                labels = batch["labels"].to(device)
                outputs = model(input_ids, attention_mask=attention_mask, labels=labels)
                loss = outputs[0]
                loss.backward()
                optim.step()



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
"""
Module that includes common tokenization functions to be applied by pandas dataframes.
"""

import itertools
import re
import string

from logai.utils import constants


def replace_delimeters(logline, delimeter_regex):
    """
    Remove customer delimeters.
    """
    return logline.replace(delimeter_regex, " ")


def tokenize(logline, config):
    """
    Common tokenization of logline and using space to separate tokens.
    """
    logline = " ".join(
        ["".join(g).strip() for k, g in itertools.groupby(logline, str.isalpha)]
    )
    logline = " ".join(
        [x for x in logline.split(" ") if len(x) > 0 and x not in string.punctuation]
    )
    return logline


def _tokenize_camel_case(logline):
    return re.sub(
        "([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", str(logline))
    ).strip()


def _tokenize_replace_digits(logline):
    tokens = []
    digits = []
    for t in logline.split():

        if t.isdigit():
            tokens.append(constants.DIGITS_SUB)
            digits.append(t)
        else:
            tokens.append(t)

    return " ".join(tokens)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import math
import torch
from torch import nn
from logai.algorithms.nn_model.forecast_nn.base_nn import (
    ForecastBasedNNParams,
    ForecastBasedNN,
)
from logai.algorithms.vectorization_algo.forecast_nn import ForecastNNVectorizedDataset
from attr import dataclass


@dataclass
class LSTMParams(ForecastBasedNNParams):
    """Config for lstm based log representation learning.

    :param num_directions: whether bidirectional or unidirectional (left to right) model.
    :param num_layers: number of hidden layers in the neural network.
    :param max_token_len:  maximum token length of the input.
    :param use_attention: whether to use attention or not.
    """

    num_directions: int = 2
    num_layers: int = 1
    max_token_len: int = None
    use_attention: bool = False


class Attention(nn.Module):
    """Attention model for lstm based log representation learning.

    :param input_size: input dimension.
    :param max_seq_len: maximum sequence length.
    """

    def __init__(self, input_size, max_seq_len):
        
        super(Attention, self).__init__()
        self.atten_w = nn.Parameter(torch.randn(max_seq_len, input_size, 1))
        self.atten_bias = nn.Parameter(torch.randn(max_seq_len, 1, 1))
        self.glorot(self.atten_w)
        self.zeros(self.atten_bias)

    def forward(self, lstm_input):
        input_tensor = lstm_input.transpose(1, 0)  # f x b x d
        input_tensor = (
            torch.bmm(input_tensor, self.atten_w) + self.atten_bias
        )  # f x b x out
        input_tensor = input_tensor.transpose(1, 0)
        atten_weight = input_tensor.tanh()

        weighted_sum = torch.bmm(atten_weight.transpose(1, 2), lstm_input).squeeze()

        return weighted_sum

    def glorot(self, tensor):
        if tensor is not None:
            stdv = math.sqrt(6.0 / (tensor.size(-2) + tensor.size(-1)))
            tensor.data.uniform_(-stdv, stdv)

    def zeros(self, tensor):
        if tensor is not None:
            tensor.data.fill_(0)


class LSTM(ForecastBasedNN):
    """LSTM based model for learning log representation through a self-supervised forecasting task over log sequences.
    :param config: parameters for lstm based model.
    """

    def __init__(self, config: LSTMParams):
        
        super().__init__(config)
        self.config = config
        self.config.model_name = "lstm"
        num_labels = self.meta_data["num_labels"]
        self.feature_type = self.config.feature_type
        self.label_type = self.config.label_type
        self.hidden_size = self.config.hidden_size
        self.num_directions = self.config.num_directions
        self.max_token_len = self.config.max_token_len
        self.use_attention = self.config.use_attention
        self.embedding_dim = self.config.embedding_dim
        self.rnn = nn.LSTM(
            input_size=self.config.embedding_dim,
            hidden_size=self.hidden_size,
            batch_first=True,
            num_layers=self.config.num_layers,
            bidirectional=(self.config.num_directions == 2),
        )
        if self.use_attention:
            assert (
                self.max_token_len is not None
            ), "max token length must be set if using attention"
            self.attn = Attention(
                self.hidden_size * self.num_directions, self.max_token_len
            )
        self.criterion = nn.CrossEntropyLoss()
        self.prediction_layer = nn.Linear(
            self.hidden_size * self.num_directions, num_labels
        )

    def forward(self, input_dict):
        """Forward method for lstm model.

        :param input_dict: dict containing the session_idx, features, window_anomalies and window_labels as in ForecastNNVectorizedDataset object.
        :return: dict containing loss and prediction tensor.
        """
        if self.label_type == "anomaly":
            y = input_dict[ForecastNNVectorizedDataset.window_anomalies].long().view(-1)
        elif self.label_type == "next_log":
            y = input_dict[ForecastNNVectorizedDataset.window_labels].long().view(-1)
        self.batch_size = y.size()[0]
        x = input_dict[ForecastNNVectorizedDataset.features]
        x = self.embedder(x)

        outputs, _ = self.rnn(x.float())
        if self.use_attention:
            representation = self.attn(outputs)
        else:
            # representation = outputs.mean(dim=1)
            representation = outputs[:, -1, :]

        logits = self.prediction_layer(representation)
        y_pred = logits.softmax(dim=-1)
        loss = self.criterion(logits, y)
        return_dict = {"loss": loss, "y_pred": y_pred}
        return return_dict



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
"""
This file is wrapping the IMB/Drain3 implementation.
Link: https://github.com/IBM/Drain3/blob/master/drain3/drain.py
"""
from typing import List, Dict

import pandas as pd
from dataclasses import dataclass
from cachetools import LRUCache, Cache

from abc import ABC, abstractmethod

from logai.algorithms.algo_interfaces import ParsingAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class DrainParams(Config):
    """Parameters for Drain Log Parser. 
    For more details on parameters see 
    https://github.com/logpai/Drain3/blob/master/drain3/drain.py.

    :param depth: The depth of tree.
    :param sim_th: The similarity threshold.
    :param max_children: The max number of children nodes.
    :param max_clusters: The max number of clusters.
    :param extra_delimiters: Extra delimiters.
    :param param_str: The wildcard parameter string.
    """
    depth: int = 3
    sim_th: float = 0.4
    max_children: int = 100
    max_clusters: int = None
    extra_delimiters: tuple = ()
    param_str: str = "*"

    @classmethod
    def from_dict(cls, config_dict):
        config = super(DrainParams, cls).from_dict(config_dict)
        if config.extra_delimiters:
            config.extra_delimiters = tuple(config.extra_delimiters)
        return config


class Profiler(ABC):
    @abstractmethod
    def start_section(self, section_name: str):
        pass

    @abstractmethod
    def end_section(self, section_name=""):
        pass

    @abstractmethod
    def report(self, period_sec=30):
        pass


class NullProfiler(Profiler):
    """A no-op profiler. Use it instead of SimpleProfiler in case you want to disable profiling."""

    def start_section(self, section_name: str):
        pass

    def end_section(self, section_name=""):
        pass

    def report(self, period_sec=30):
        pass


class LogCluster:

    __slots__ = ["log_template_tokens", "cluster_id", "size"]

    def __init__(self, log_template_tokens: list, cluster_id: int):
        self.log_template_tokens = tuple(log_template_tokens)
        self.cluster_id = cluster_id
        self.size = 1

    def get_template(self):
        return " ".join(self.log_template_tokens)

    def __str__(self):
        return f"ID={str(self.cluster_id).ljust(5)} : size={str(self.size).ljust(10)}: {self.get_template()}"


class LogClusterCache(LRUCache):
    """
    Least Recently Used (LRU) cache which allows callers to conditionally skip
    cache eviction algorithm when accessing elements.
    """

    def __missing__(self, key):
        return None

    def get(self, key):
        """
        Returns the value of the item with the specified key without updating
        the cache eviction algorithm.
        """
        return Cache.__getitem__(self, key)


class Node:
    __slots__ = ["key_to_child_node", "cluster_ids"]

    def __init__(self):
        self.key_to_child_node: Dict[str, Node] = {}
        self.cluster_ids: List[int] = []


@factory.register("parsing", "drain", DrainParams)
class Drain(ParsingAlgo):
    def __init__(self, params: DrainParams, profiler=NullProfiler()):
        if params.depth < 3:
            raise ValueError("depth argument must be at least 3")

        self.log_cluster_depth = params.depth
        self.max_node_depth = (
            params.depth - 2
        )  # max depth of a prefix tree node, starting from zero
        self.sim_th = params.sim_th
        self.max_children = params.max_children
        self.root_node = Node()
        self.profiler = profiler
        self.extra_delimiters = params.extra_delimiters
        self.max_clusters = params.max_clusters
        self.param_str = params.param_str

        # key: int, value: LogCluster
        self.id_to_cluster = (
            {}
            if params.max_clusters is None
            else LogClusterCache(maxsize=params.max_clusters)
        )
        self.clusters_counter = 0

    @property
    def clusters(self):
        return self.id_to_cluster.values()

    @staticmethod
    def has_numbers(s):
        return any(char.isdigit() for char in s)

    def _tree_search(
        self, root_node: Node, tokens: list, sim_th: float, include_params: bool
    ):

        # at first level, children are grouped by token (word) count
        token_count = len(tokens)
        cur_node = root_node.key_to_child_node.get(str(token_count))

        # no template with same token count yet
        if cur_node is None:
            return None

        # handle case of empty log string - return the single cluster in that group
        if token_count == 0:
            return self.id_to_cluster.get(cur_node.cluster_ids[0])

        # find the leaf node for this log - a path of nodes matching the first N tokens (N=tree depth)
        cur_node_depth = 1
        for token in tokens:
            # at max depth
            if cur_node_depth >= self.max_node_depth:
                break

            # this is last token
            if cur_node_depth == token_count:
                break

            key_to_child_node = cur_node.key_to_child_node
            cur_node = key_to_child_node.get(token)
            if cur_node is None:  # no exact next token exist, try wildcard node
                cur_node = key_to_child_node.get(self.param_str)
            if cur_node is None:  # no wildcard node exist
                return None

            cur_node_depth += 1

        # get best match among all clusters with same prefix, or None if no match is above sim_th
        cluster = self._fast_match(cur_node.cluster_ids, tokens, sim_th, include_params)
        return cluster

    def _add_seq_to_prefix_tree(self, root_node, cluster: LogCluster):
        token_count = len(cluster.log_template_tokens)
        token_count_str = str(token_count)
        if token_count_str not in root_node.key_to_child_node:
            first_layer_node = Node()
            root_node.key_to_child_node[token_count_str] = first_layer_node
        else:
            first_layer_node = root_node.key_to_child_node[token_count_str]

        cur_node = first_layer_node

        # handle case of empty log string
        if token_count == 0:
            cur_node.cluster_ids = [cluster.cluster_id]
            return

        current_depth = 1
        for token in cluster.log_template_tokens:

            # if at max depth or this is last token in template - add current log cluster to the leaf node
            if current_depth >= self.max_node_depth or current_depth >= token_count:
                # clean up stale clusters before adding a new one.
                new_cluster_ids = []
                for cluster_id in cur_node.cluster_ids:
                    if cluster_id in self.id_to_cluster:
                        new_cluster_ids.append(cluster_id)
                new_cluster_ids.append(cluster.cluster_id)
                cur_node.cluster_ids = new_cluster_ids
                break

            # if token not matched in this layer of existing tree.
            if token not in cur_node.key_to_child_node:
                if not self.has_numbers(token):
                    if self.param_str in cur_node.key_to_child_node:
                        if len(cur_node.key_to_child_node) < self.max_children:
                            new_node = Node()
                            cur_node.key_to_child_node[token] = new_node
                            cur_node = new_node
                        else:
                            cur_node = cur_node.key_to_child_node[self.param_str]
                    else:
                        if len(cur_node.key_to_child_node) + 1 < self.max_children:
                            new_node = Node()
                            cur_node.key_to_child_node[token] = new_node
                            cur_node = new_node
                        elif len(cur_node.key_to_child_node) + 1 == self.max_children:
                            new_node = Node()
                            cur_node.key_to_child_node[self.param_str] = new_node
                            cur_node = new_node
                        else:
                            cur_node = cur_node.key_to_child_node[self.param_str]

                else:
                    if self.param_str not in cur_node.key_to_child_node:
                        new_node = Node()
                        cur_node.key_to_child_node[self.param_str] = new_node
                        cur_node = new_node
                    else:
                        cur_node = cur_node.key_to_child_node[self.param_str]

            # if the token is matched
            else:
                cur_node = cur_node.key_to_child_node[token]

            current_depth += 1

    # seq1 is template
    def _get_seq_distance(self, seq1, seq2, include_params: bool):
        assert len(seq1) == len(seq2)
        sim_tokens = 0
        param_count = 0

        for token1, token2 in zip(seq1, seq2):
            if token1 == self.param_str:
                param_count += 1
                continue
            if token1 == token2:
                sim_tokens += 1

        if include_params:
            sim_tokens += param_count

        ret_val = float(sim_tokens) / len(seq1)

        return ret_val, param_count

    def _fast_match(
        self, cluster_ids: list, tokens: list, sim_th: float, include_params: bool
    ):
        """
        Find the best match for a log message (represented as tokens) versus a list of clusters.

        :param cluster_ids: List of clusters to match against (represented by their IDs).
        :param tokens: the log message, separated to tokens.
        :param sim_th: minimum required similarity threshold (None will be returned in no clusters reached it).
        :param include_params: consider tokens matched to wildcard parameters in similarity threshold.
        :return: Best match cluster or None.
        """
        match_cluster = None

        max_sim = -1
        max_param_count = -1
        max_cluster = None

        for cluster_id in cluster_ids:
            # Try to retrieve cluster from cache with bypassing eviction
            # algorithm as we are only testing candidates for a match.
            cluster = self.id_to_cluster.get(cluster_id)
            if cluster is None:
                continue
            cur_sim, param_count = self._get_seq_distance(
                cluster.log_template_tokens, tokens, include_params
            )
            if cur_sim > max_sim or (
                cur_sim == max_sim and param_count > max_param_count
            ):
                max_sim = cur_sim
                max_param_count = param_count
                max_cluster = cluster

        if max_sim >= sim_th:
            match_cluster = max_cluster

        return match_cluster

    def _create_template(self, seq1, seq2):
        assert len(seq1) == len(seq2)
        ret_val = list(seq2)

        for i, (token1, token2) in enumerate(zip(seq1, seq2)):
            if token1 != token2:
                ret_val[i] = self.param_str

        return ret_val

    def _print_tree(self, file=None, max_clusters=5):
        self._print_node("root", self.root_node, 0, file, max_clusters)

    def _print_node(self, token, node, depth, file, max_clusters):
        out_str = "\t" * depth

        if depth == 0:
            out_str += f"<{token}>"
        elif depth == 1:
            out_str += f"<L={token}>"
        else:
            out_str += f'"{token}"'

        if len(node.cluster_ids) > 0:
            out_str += f" (cluster_count={len(node.cluster_ids)})"

        print(out_str, file=file)

        for token, child in node.key_to_child_node.items():
            self._print_node(token, child, depth + 1, file, max_clusters)

        for cid in node.cluster_ids[:max_clusters]:
            cluster = self.id_to_cluster[cid]
            out_str = "\t" * (depth + 1) + str(cluster)
            print(out_str, file=file)

    def _get_content_as_tokens(self, content):
        content = content.strip()
        for delimiter in self.extra_delimiters:
            content = content.replace(delimiter, " ")
        content_tokens = content.split()
        return content_tokens

    def _add_log_message(self, content: str):
        content_tokens = self._get_content_as_tokens(content)

        if self.profiler:
            self.profiler.start_section("tree_search")
        match_cluster = self._tree_search(
            self.root_node, content_tokens, self.sim_th, False
        )
        if self.profiler:
            self.profiler.end_section()

        # Match no existing log cluster
        if match_cluster is None:
            if self.profiler:
                self.profiler.start_section("create_cluster")
            self.clusters_counter += 1
            cluster_id = self.clusters_counter
            match_cluster = LogCluster(content_tokens, cluster_id)
            self.id_to_cluster[cluster_id] = match_cluster
            self._add_seq_to_prefix_tree(self.root_node, match_cluster)
            update_type = "cluster_created"

        # Add the new log message to the existing cluster
        else:
            if self.profiler:
                self.profiler.start_section("cluster_exist")
            new_template_tokens = self._create_template(
                content_tokens, match_cluster.log_template_tokens
            )
            if tuple(new_template_tokens) == match_cluster.log_template_tokens:
                update_type = "none"
            else:
                match_cluster.log_template_tokens = tuple(new_template_tokens)
                update_type = "cluster_template_changed"
            match_cluster.size += 1
            # Touch cluster to update its state in the cache.
            # noinspection PyStatementEffect
            self.id_to_cluster[match_cluster.cluster_id]

        if self.profiler:
            self.profiler.end_section()

        return match_cluster, update_type

    def match(self, content: str):
        """
        Match against an already existing cluster. Match shall be perfect (sim_th=1.0).
        New cluster will not be created as a result of this call, nor any cluster modifications.

        :param content: The log message to match.
        :return: Matched cluster or None of no match found.
        """
        content_tokens = self._get_content_as_tokens(content)
        match_cluster = self._tree_search(self.root_node, content_tokens, 1.0, True)
        return match_cluster

    def _get_total_cluster_size(self):
        return self.clusters_counter

    def fit(self, logline: pd.Series):
        for l in logline:
            if not isinstance(l, str):
                continue
            self._add_log_message(l)

    def parse(self, logline: pd.Series) -> pd.Series:
        """Parse method to run log parser on a given log data.

        :param logline: The raw log data to be parsed.
        :returns: The parsed log data.
        """
        self.fit(logline)
        parsed_logline = []
        for line in logline:
            parsed_logline.append(" ".join(self.match(line).log_template_tokens))
        return pd.Series(parsed_logline, index=logline.index)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass

from sklearn.ensemble import IsolationForest

from logai.algorithms.algo_interfaces import AnomalyDetectionAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class IsolationForestParams(Config):
    """Parameters for isolation forest based anomaly detection. For more explanation of the parameters see the documentation page
    in https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html.

    :param n_estimators: The number of base estimators in the ensemble.
    :param max_samples: The number of samples to draw from X to train each base estimator.
    :param contamination: The amount of contamination of the data set, i.e. the proportion of outliers in the data set.
    :param max_features: The number of features to draw from X to train each base estimator.
    :param bootstrap: If True, individual trees are fit on random subsets of the training data sampled with
        replacement. If False, sampling without replacement is performed.
    :param n_jobs: The number of jobs to run in parallel for both fit and predict.
    :param random_state: Controls the pseudo-randomness of the selection of the feature and split values
        for each branching step and each tree in the forest.
    :param verbose: Controls the verbosity of the tree building process.
    :param warm_start: When set to True, reuse the solution of the previous call to fit and add more estimators
        to the ensemble, otherwise, just fit a whole new forest.
    """
    n_estimators: int = 100
    max_samples: str = "auto"
    contamination: str = "auto"
    max_features: float = 1.0
    bootstrap: bool = False
    n_jobs: int = None
    random_state: object = None
    verbose: int = 0
    warm_start: bool = False


@factory.register("detection", "isolation_forest", IsolationForestParams)
class IsolationForestDetector(AnomalyDetectionAlgo):
    """Isolation Forest based Anomaly Detector. This is a wrapper for the Isolation forest in scikit-learn library.
    """
    def __init__(self, params: IsolationForestParams):
        """Constructor for isolation forest based anomaly detector.
        
        :param params: An object of IsolationForestParams containing parameters of Isolation Forest.
        """
        self.model = IsolationForest(
            n_estimators=params.n_estimators,
            max_samples=params.max_samples,
            contamination=params.contamination,
            max_features=params.max_features,
            bootstrap=params.bootstrap,
            n_jobs=params.n_jobs,
            random_state=params.random_state,
            verbose=params.verbose,
            warm_start=params.verbose,
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Fits an isolation forest model.

        :param log_features: The input for model training.
        :return: The scores of the training dataset.
        """
        self.model.fit(log_features)
        train_scores = self.model.score_samples(log_features)
        train_scores = pd.DataFrame(train_scores, index=log_features.index)
        train_scores["trainval"] = True
        return train_scores

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """
        Predicts anomalies.

        :param log_features: The input for inference.
        :return: A pandas dataframe of the predicted anomaly scores.
        """
        test_scores = self.model.predict(log_features)
        test_scores = pd.DataFrame(
            pd.Series(test_scores, index=log_features.index, name="anom_score")
        )

        test_scores["trainval"] = False
        return test_scores



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import torch
import random
import os
import numpy as np


def tensor2flatten_arr(tensor):
    """Convert tensor to flat numpy array.
    
    :param tensor: tensor object
    :return: flat numpy array
    """
    return tensor.data.cpu().numpy().reshape(-1)


def seed_everything(seed=1234):
    """Fix the random seeds throughout the python environment

    :param seed: Seed value. Defaults to 1234.
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def set_device(gpu: int = None):
    """Set device (cpu or gpu). Use -1 to specify cpu.
    If not manually set device would be automatically set to gpu.
    if gpu is available otherwise cpu would be used.
    
    :param gpu: device number of gpu (use -1 for cpu). Defaults to None.
    :return:  torch device type object.
    """
    if not gpu:
        if torch.cuda.is_available():
            device = torch.device("cuda:" + str(torch.cuda.current_device()))
        else:
            device = torch.device("cpu")
    else:
        device = torch.device("cuda:" + str(gpu))
    return device



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import logging

import pandas as pd

from logai.applications.application_interfaces import WorkFlowConfig
from logai.dataloader.data_loader import FileDataLoader
from logai.dataloader.data_model import LogRecordObject
from logai.dataloader.openset_data_loader import OpenSetDataLoader
from logai.information_extraction.log_parser import LogParser
from logai.preprocess.preprocessor import Preprocessor
from logai.utils import constants


class AutoLogSummarization:
    """
    The unified framework for log parsing analysis.
    How to use, the design of this analysis app should follow the general workflow of
    automated log parsing. The workflow should be able to control solely by `WorkFlowConfig`.

    :param config: The configuration that controls the behavior this app.
    """

    def __init__(self, config: WorkFlowConfig):
        
        self.config = config
        self._parsing_results = pd.DataFrame()
        self._attributes = None
        self._timestamp = None
        pass

    @property
    def parsing_results(self):
        return self._parsing_results

    @property
    def attributes(self):
        return self._attributes

    @property
    def log_patterns(self):
        if self._parsing_results.empty:
            return None
        return self._parsing_results[constants.PARSED_LOGLINE_NAME].unique()

    def get_parameter_list(self, log_pattern):
        """
        For a given log pattern, return the dynamic parameters.
        
        :param log_pattern: The input log pattern.
        :return: The parameter list with Values, valuecounts and position.
        """
        para_list = pd.DataFrame(None, columns=["position", "value_counts", "values"])
        if self._parsing_results.empty or not log_pattern:
            return para_list

        res = self._parsing_results
        parameters = res.loc[res[constants.PARSED_LOGLINE_NAME] == log_pattern][
            constants.PARAMETER_LIST_NAME
        ]

        para_list["values"] = pd.Series(
            pd.DataFrame(parameters.tolist()).T.values.tolist()
        )
        para_list["position"] = [
            "POSITION_{}".format(v) for v in para_list.index.values
        ]
        para_list["value_counts"] = [
            len(list(filter(None, v))) for v in para_list["values"]
        ]
        return para_list

    def recognize_parameter_entity(self, para_list):
        """
        Placeholder for log parameter entity recognization
        """
        pass

    def summarize_numeric_paramters(self, paras: list):
        """
        Placeholder for numeric parameter summarization
        """
        pass

    def find_log_pattern(self, logline: str, return_para_list: bool = True):
        """
        Find the log pattern for a given logline, return all dynamic parameters in this log pattern if needed.
        """
        log_pattern = None
        para_list = None
        if not self._parsing_results.empty:
            res = self._parsing_results.loc[
                self._parsing_results[constants.LOGLINE_NAME] == logline
            ]
            log_patterns = res[constants.PARSED_LOGLINE_NAME]
            if len(log_patterns) == 0:
                return None
            if len(log_patterns) == 1:
                log_pattern_index = log_patterns.index[0]
                log_pattern = log_patterns.iloc[log_pattern_index]
            else:
                logging.warning("multiple log pa†terns are found!")
                log_pattern_index = log_patterns.index[0]
                log_pattern = log_patterns.iloc[log_pattern_index]
            if return_para_list:
                para_list = self.get_parameter_list(log_pattern)

        return log_pattern, para_list

    def execute(self):
        """
        Execute auto log parsing analysis. Store the results and index for searching.
        """
        # load data
        logrecord = self._load_data()

        if not logrecord.attributes.empty:
            self._attributes = logrecord.attributes.astype(str)

        if not logrecord.timestamp.empty:
            self._timestamp = logrecord.timestamp

        # preprocess
        preprocessed_loglines = self._preprocess(logrecord)

        parser = LogParser(self.config.log_parser_config)
        parsed_results = parser.parse(preprocessed_loglines.dropna())
        self._parsing_results = parsed_results[
            [constants.PARSED_LOGLINE_NAME, constants.PARAMETER_LIST_NAME]
        ]

        self._parsing_results = self._parsing_results.join(logrecord.body)

        if self._attributes is not None:
            self._parsing_results = self._parsing_results.join(self._attributes)

        if self._timestamp is not None:
            self._parsing_results = self._parsing_results.join(self._timestamp)

        return

    def _load_data(self):
        if self.config.open_set_data_loader_config is not None:
            dataloader = OpenSetDataLoader(self.config.open_set_data_loader_config)
            logrecord = dataloader.load_data()
        elif self.config.data_loader_config is not None:
            dataloader = FileDataLoader(self.config.data_loader_config)
            logrecord = dataloader.load_data()
        else:
            raise ValueError(
                "data_loader_config or open_set_data_loader_config is needed to load data."
            )
        return logrecord

    def _preprocess(self, logrecord: LogRecordObject):
        logline = logrecord.body[constants.LOGLINE_NAME]

        # Preprocessor cleans the loglines
        preprocessor = Preprocessor(self.config.preprocessor_config)
        preprocessed_loglines, _ = preprocessor.clean_log(logline)

        return preprocessed_loglines



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from attr import dataclass
from sklearn.preprocessing import OrdinalEncoder

from logai.algorithms.algo_interfaces import CategoricalEncodingAlgo
from logai.config_interfaces import Config


@dataclass
class OrdinalEncodingParams(Config):
    """
    Configuration of OrdinalEncoding. For more details on the parameters see 
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OrdinalEncoder.html.

    :param categories: Categories (unique values) per feature.
    :param dtype: Desired dtype of output.
    :param handle_unknown: Specifies the way unknown categories are handled during transform.
    :param unknown_value: When the parameter handle_unknown is set to ‘use_encoded_value’,
        this parameter is required and will set the encoded value of unknown categories.
    """
    categories: str = "auto"
    dtype: np.float64 = np.float64
    handle_unknown: str = "error"
    unknown_value: object = None


class OrdinalEncoding(CategoricalEncodingAlgo):
    """
    Implementation of ordinal encoder. This is a wrapper class for the OrdinalEncoder from scikit learn 
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OrdinalEncoder.html.
    """

    def __init__(self, params: OrdinalEncodingParams):
        self.model = OrdinalEncoder(
            categories=params.categories,
            dtype=params.dtype,
            handle_unknown=params.handle_unknown,
            unknown_value=params.unknown_value,
        )

    def fit_transform(self, log_attributes: pd.DataFrame) -> pd.DataFrame:
        """
        Fits and transforms log attributes into ordinal encoding categories.
        
        :param log_attributes: A list of log attributes in text format.
        :return: A pandas dataframe of ordinal encoding categories.
        """
        self.model.fit(log_attributes)
        res_column_names = ["{}-categorical".format(c) for c in log_attributes.columns]
        res = pd.DataFrame(
            self.model.transform(log_attributes), index=log_attributes.index
        )
        res.columns = res_column_names
        return res



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass
from sklearn.cluster import DBSCAN

from logai.algorithms.algo_interfaces import ClusteringAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class DbScanParams(Config):
    """Parameters for DBScan based clustering algorithm. For more details on parameters see 
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html.

    :param eps: The maximum distance between two samples for one to be considered as in the
        neighborhood of the other.
    :param min_samples: The number of samples (or total weight) in a neighborhood for a point
        to be considered as a core point.
    :param metric: The metric to use when calculating distance between instances in a feature array.
    :param metric_params: Additional keyword arguments for the metric function.
    :param algorithm: The algorithm to be used by the NearestNeighbors module to compute pointwise
        distances and find nearest neighbors, i.e., ``{‘auto’, ‘ball_tree’, ‘kd_tree’, ‘brute’}``.
    :param leaf_size: Leaf size passed to BallTree or cKDTree.
    :param p: The power of the Minkowski metric to be used to calculate distance between points.
    :param n_jobs: The number of parallel jobs to run.
    """
    eps: float = 0.3
    min_samples: int = 10
    metric: str = "euclidean"
    metric_params: object = None
    algorithm: str = "auto"
    leaf_size: int = 30
    p: float = None
    n_jobs: int = None


@factory.register("clustering", "dbscan", DbScanParams)
class DbScanAlgo(ClusteringAlgo):
    """
    DBSCAN algorithm for log clustering. This is a wrapper class for DBScan based from scikit-learn library 
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    """

    def __init__(self, params: DbScanParams):
        self.model = DBSCAN(
            eps=params.eps,
            min_samples=params.min_samples,
            metric=params.metric,
            metric_params=params.metric_params,
            algorithm=params.algorithm,
            leaf_size=params.leaf_size,
            p=params.p,
            n_jobs=params.n_jobs,
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Trains a DBSCAN model.

        :param log_features: The log features as training data.
        """
        self.model.fit(log_features)

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """
        Predicts using the trained DBSCAN model.

        :param log_features: The log features for inference.
        :return: A pandas series of cluster labels.
        """
        res = self.model.fit_predict(log_features)
        return pd.Series(res, index=log_features.index)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from attr import dataclass
from sklearn.feature_extraction.text import TfidfVectorizer

from logai.algorithms.algo_interfaces import VectorizationAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class TfIdfParams(Config):
    """
    Configuration of TF-IDF vectorizer. For more details of parameters see 
    https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html.

    :param input: ``{'filename', 'file', 'content'}``; If `'filename'`, the sequence passed as an argument to
        fit is expected to be a list of filenames that need reading to fetch the raw content to analyze.
    :param encoding: If bytes or files are given to analyze, this encoding is used to decode.
    :param decode_error: ``{'strict', 'ignore', 'replace'}``;
        Instruction on what to do if a byte sequence is given to analyze that
        contains characters not of the given `encoding`. By default, it is
        'strict', meaning that a UnicodeDecodeError will be raised. Other
        values are 'ignore' and 'replace'.
    :param strip_accents: Remove accents and perform other character normalization
        during the preprocessing step.
    :param lowercase: Convert all characters to lowercase before tokenizing.
    :param preprocessor: Override the preprocessing (string transformation) stage while
        preserving the tokenizing and n-grams generation steps.
    :param tokenizer: Override the string tokenization step while preserving the
        preprocessing and n-grams generation steps.
    :param analyzer: Whether the feature should be made of word or character n-grams.
    :param stop_words: If a string, it is passed to _check_stop_list and the appropriate stop
        list is returned. 'english' is currently the only supported string value.
    :param token_pattern: Regular expression denoting what constitutes a "token", only used
        if ``analyzer == 'word'``.
    :param ngram_range: The lower and upper boundary of the range of n-values for different
        n-grams to be extracted.
    :param max_df: When building the vocabulary ignore terms that have a document
        frequency strictly higher than the given threshold.
    :param min_df: When building the vocabulary ignore terms that have a document
        frequency strictly lower than the given threshold.
    :param max_features: If not None, build a vocabulary that only consider the top
        max_features ordered by term frequency across the corpus.
    :param vocabulary: Either a Mapping (e.g., a dict) where keys are terms and values are
        indices in the feature matrix, or an iterable over terms. If not
        given, a vocabulary is determined from the input documents.
    :param binary: If True, all non-zero term counts are set to 1. This does not mean
        outputs will have only 0/1 values, only that the tf term in tf-idf
        is binary.
    :param dtype: Type of the matrix returned by fit_transform() or transform().
    :param norm: Each output row will have unit norm, i.e., ``{'l1', 'l2'}``.
    :param use_idf: Enable inverse-document-frequency reweighting. If False, idf(t) = 1.
    :param smooth_idf: Smooth idf weights by adding one to document frequencies.
    :param sublinear_tf: Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).
    """

    input: str = "content"
    encoding: str = "utf-8"
    decode_error: str = "strict"
    strip_accents: object = None
    lowercase: bool = True
    preprocessor: object = None
    tokenizer: object = None
    analyzer: str = "word"
    stop_words: object = None
    token_pattern: str = r"(?u)\b\w\w+\b"
    ngram_range: tuple = (1, 1)
    max_df: float = 1.0
    min_df: int = 1
    max_features: object = None
    vocabulary: object = None
    binary: bool = False
    dtype: object = np.float64
    norm: str = "l2"
    use_idf: bool = True
    smooth_idf: bool = True
    sublinear_tf: bool = False


@factory.register("vectorization", "tfidf", TfIdfParams)
class TfIdf(VectorizationAlgo):
    """
    TfIdf based vectorizer for log data. This is a wrapper class of the TF-IDF Vectorizer algorithm from scikit-learn.
    https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html.
    """

    def __init__(self, params: TfIdfParams, **kwargs):
        """
        Initializes TF-IDF vectorizer.

        :param params: TF-IDF algorithm parameters.
        :param kwargs: Optional k-v based params.
        """
        self.model = TfidfVectorizer(
            input=params.input,
            encoding=params.encoding,
            decode_error=params.decode_error,
            strip_accents=params.strip_accents,
            lowercase=params.lowercase,
            preprocessor=params.preprocessor,
            tokenizer=params.tokenizer,
            analyzer=params.analyzer,
            stop_words=params.stop_words,
            token_pattern=params.token_pattern,
            ngram_range=params.ngram_range,
            max_df=params.max_df,
            min_df=params.min_df,
            max_features=params.max_features,
            vocabulary=params.vocabulary,
            binary=params.binary,
            dtype=params.dtype,
            norm=params.norm,
            use_idf=params.use_idf,
            smooth_idf=params.smooth_idf,
            sublinear_tf=params.sublinear_tf,
        )

    def fit(self, loglines: pd.Series):
        """
        Trains a TF-IDF model.

        :param loglines: The input training dataset.
        """
        self.model.fit(loglines)
        self.vocab = self.model.vocabulary_
        self.vocab_size = len(self.vocab)

    def transform(self, loglines: pd.Series) -> pd.Series:
        """
        Transforms loglines into log vectors.

        :param loglines: The input test dataset.
        :return: The transformed log vectors.
        """
        res = self.model.transform(loglines)
        return pd.Series(res.todense().tolist(), index=loglines.index).apply(
            lambda x: np.array(x)
        )

    def summary(self):
        """
        Generates model summary.
        """
        return self.model.summary()



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd

from logai.algorithms.clustering_algo.dbscan import DBSCAN
from logai.algorithms.clustering_algo.kmeans import KMeans
from logai.analysis.clustering import Clustering
from logai.applications.application_interfaces import WorkFlowConfig
from logai.dataloader.data_loader import FileDataLoader
from logai.dataloader.openset_data_loader import OpenSetDataLoader
from logai.information_extraction.categorical_encoder import CategoricalEncoder
from logai.information_extraction.feature_extractor import FeatureExtractor
from logai.information_extraction.log_parser import LogParser
from logai.information_extraction.log_vectorizer import LogVectorizer
from logai.preprocess.preprocessor import Preprocessor
from logai.utils import constants
from logai.utils.functions import pad


class LogClustering:
    """
    Clustering Application class defines log clustering workflow.
    It includes which algorithm to use.
    """

    def __init__(self, config):
        self.config = config
        self._loglines = pd.DataFrame()
        self._timestamps = pd.DataFrame()
        self._attributes = pd.DataFrame()
        self._feature_df = pd.DataFrame()
        self._clusters = pd.DataFrame()
        self.MAX_LEN = 100
        pass

    @property
    def clusters(self):
        return self._clusters

    @property
    def attributes(self):
        return self._attributes

    @property
    def timestamps(self):
        return self._timestamps

    @property
    def event_index(self):
        return self._index

    @property
    def loglines(self):
        return self._loglines

    @property
    def logline_with_clusters(self):
        event_cluster_df = pd.concat(
            (self.clusters, self.loglines, self.attributes, self.timestamps), axis=1
        )
        return event_cluster_df

    def execute(self):

        # Load data
        logrecord = self._load_data()

        # Preprocessor cleans the loglines
        logline = logrecord.body[constants.LOGLINE_NAME]

        self._loglines = logline
        self._timestamps = logrecord.timestamp

        preprocessor = Preprocessor(self.config.preprocessor_config)
        preprocessed_loglines, _ = preprocessor.clean_log(logline)

        # Parsing
        parser = LogParser(self.config.log_parser_config)
        parsed_results = parser.parse(preprocessed_loglines.dropna())

        parsed_loglines = parsed_results[constants.PARSED_LOGLINE_NAME]

        # Vectorization
        vectorizor = LogVectorizer(self.config.log_vectorizer_config)
        vectorizor.fit(parsed_loglines)

        # Log vector is a pandas.Series
        log_vectors = vectorizor.transform(parsed_loglines)

        # Categorical Encoding
        encoder = CategoricalEncoder(self.config.categorical_encoder_config)

        self._attributes = logrecord.attributes.astype(str)

        attributes = encoder.fit_transform(logrecord.attributes)
        attributes.columns = logrecord.attributes.columns

        padded_log_vectors = log_vectors.apply(pad, max_len=self.MAX_LEN)

        feature_for_clustering = pd.DataFrame(
            [vec for vec in padded_log_vectors], index=padded_log_vectors.index
        )

        feature_for_clustering = feature_for_clustering.join(attributes)

        log_clustering = Clustering(self.config.clustering_config)

        log_clustering.fit(feature_for_clustering)

        self._clusters = (
            log_clustering.predict(feature_for_clustering)
            .astype(str)
            .rename("cluster_id")
        )
        self._index = self._clusters.index

        return

    def _load_data(self):
        if self.config.open_set_data_loader_config is not None:
            dataloader = OpenSetDataLoader(self.config.open_set_data_loader_config)
            logrecord = dataloader.load_data()
        elif self.config.data_loader_config is not None:
            dataloader = FileDataLoader(self.config.data_loader_config)
            logrecord = dataloader.load_data()
        else:
            raise ValueError(
                "data_loader_config or open_set_data_loader_config is needed to load data."
            )
        return logrecord



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from .dbl import DBLDetector
from .distribution_divergence import DistributionDivergence
from .ets import ETSDetector
from .isolation_forest import IsolationForestDetector
from .local_outlier_factor import LOFDetector
from .one_class_svm import OneClassSVMDetector
from logai.utils.misc import is_torch_available, \
    is_transformers_available

_MODULES = [
    "DBLDetector",
    "DistributionDivergence",
    "ETSDetector",
    "IsolationForestDetector",
    "LOFDetector",
    "OneClassSVMDetector"
]

if is_torch_available() and is_transformers_available():
    from .forecast_nn import ForecastBasedLSTM, ForecastBasedCNN, ForecastBasedTransformer
    from .logbert import LogBERT

    _MODULES += [
        "LogBERT",
        "ForecastBasedLSTM",
        "ForecastBasedCNN",
        "ForecastBasedTransformer"
    ]

__all__ = _MODULES



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from attr import dataclass

from logai.algorithms.algo_interfaces import ParsingAlgo
from logai.config_interfaces import Config

"""
** Need to refactor the implementaiton of IPLoM **
This is wrapping the logpai/logparser implementation of AEL algorithm
link: https://github.com/logpai/logparser/blob/master/logparser/IPLoM/IPLoM.py
"""

import copy
import sys
import re
import pandas as pd
import hashlib

from logai.algorithms.factory import factory


class Partition:
    """Wrap around the logs and the step number"""

    def __init__(self, stepNo, numOfLogs=0, lenOfLogs=0):
        self.logLL = []
        self.stepNo = stepNo
        self.valid = True
        self.numOfLogs = numOfLogs
        self.lenOfLogs = lenOfLogs


class Event:
    """Event class to wrap a Log Event 
    """
    def __init__(self, eventStr):
        self.eventStr = eventStr
        self.eventId = hashlib.md5(" ".join(eventStr).encode("utf-8")).hexdigest()[0:8]
        self.eventCount = 0


@dataclass
class IPLoMParams(Config):
    """Parameters for the IPLoM Log Parser. For more details on parameters see 
    https://github.com/logpai/logparser/tree/master/logparser/IPLoM.

    :param rex: The rex string.
    :param logformat: The log format.
    :param maxEventLen: The max event length.
    :param step2Support: The step to support.
    :param lowerBound: The lower bound threshold.
    :param upperBound: The upper bound threshold.
    :param keep_para: Whether to keep parameters.
    """
    rex: str = None
    logformat: str = None
    maxEventLen: int = 200
    step2Support: float = 0
    PST: float = 0
    CT: float = 0
    lowerBound: float = 0.25
    upperBound: float = 0.9
    keep_para: bool = True


@factory.register("parsing", "iplom", IPLoMParams)
class IPLoM(ParsingAlgo):
    """IPLoM Log Parsing algorithm. For details see 
    https://github.com/logpai/logparser/tree/master/logparser/IPLoM.
    """
    def __init__(self, params: IPLoMParams):
        self.para = params
        self.partitionsL = []
        self.eventsL = []
        self.output = []
        self.keep_para = params.keep_para

        # Initialize some partitions which contain logs with different length
        for logLen in range(self.para.maxEventLen + 1):
            self.partitionsL.append(Partition(stepNo=1, numOfLogs=0, lenOfLogs=logLen))

    def fit(self, loglines: pd.Series):
        """
        IPLoM does not support model fit. Call parse() directly with given input and get log templates.
        """
        pass

    def parse(self, loglines: pd.Series) -> pd.Series:
        """Parsing method to parse the raw log data.

        :param loglines: The raw log data.
        :returns: The parsed log data.
        """
        self._Step1(loglines)
        self._Step2()
        self._Step3()
        self._Step4()
        self._getOutput()
        eventID_template = {
            event.eventId: " ".join(event.eventStr) for event in self.eventsL
        }

        self.output.sort(key=lambda x: int(x[0]))
        res = pd.Series(
            [eventID_template[logL[1]] for logL in self.output], index=loglines.index
        )
        return res

    def _Step1(self, loglines: pd.Series):
        self.df_log = pd.DataFrame(loglines)
        lineCount = 1
        for idx, line in self.df_log.iterrows():
            line = line[loglines.name]
            # If line is empty, skip
            if line.strip() == "":
                continue

            if self.para.rex:
                for currentRex in self.para.rex:
                    line = re.sub(currentRex, "", line)

            wordSeq = list(filter(lambda x: x != "", re.split(r"[\s=:,]", line)))
            if not wordSeq:
                wordSeq = [" "]

            # Generate terms list, with ID in the end
            wordSeq.append(str(lineCount))
            lineCount += 1

            # Add current log to the corresponding partition
            self.partitionsL[len(wordSeq) - 1].logLL.append(wordSeq)
            self.partitionsL[len(wordSeq) - 1].numOfLogs += 1

        for partition in self.partitionsL:
            if partition.numOfLogs == 0:
                partition.valid = False

            elif (
                self.para.PST != 0
                and 1.0 * partition.numOfLogs / lineCount < self.para.PST
            ):
                for logL in partition.logLL:
                    self.partitionsL[0].logLL.append(logL)
                    self.partitionsL[0].numOfLogs += 1
                partition.valid = False

    def _Step2(self):

        for partition in self.partitionsL:

            if not partition.valid:
                continue

            if partition.numOfLogs <= self.para.step2Support:
                continue

            # Avoid going through newly generated partitions
            if partition.stepNo == 2:
                break

            # For each column, create a set to hold the unique tokens in that column.
            # And finally, calculate the number of the unique tokens in each column
            uniqueTokensCountLS = []
            for columnIdx in range(partition.lenOfLogs):
                uniqueTokensCountLS.append(set())

            for logL in partition.logLL:
                for columnIdx in range(partition.lenOfLogs):
                    uniqueTokensCountLS[columnIdx].add(logL[columnIdx])

            # Find the column with minimum unique tokens
            minColumnIdx = 0
            minColumnCount = len(uniqueTokensCountLS[0])

            for columnIdx in range(partition.lenOfLogs):
                if minColumnCount > len(uniqueTokensCountLS[columnIdx]):
                    minColumnCount = len(uniqueTokensCountLS[columnIdx])
                    minColumnIdx = columnIdx

            # If there is one column with one unique term, do not split this partition
            if minColumnCount == 1:
                continue

            # From split-token to log list
            logDLL = {}
            for logL in partition.logLL:
                if logL[minColumnIdx] not in logDLL:
                    logDLL[logL[minColumnIdx]] = []
                logDLL[logL[minColumnIdx]].append(logL)

            for key in logDLL:
                if (
                    self.para.PST != 0
                    and 1.0 * len(logDLL[key]) / partition.numOfLogs < self.para.PST
                ):
                    self.partitionsL[0].logLL += logDLL[key]
                    self.partitionsL[0].numOfLogs += len(logDLL[key])
                else:
                    newPartition = Partition(
                        stepNo=2,
                        numOfLogs=len(logDLL[key]),
                        lenOfLogs=partition.lenOfLogs,
                    )
                    newPartition.logLL = logDLL[key]
                    self.partitionsL.append(newPartition)

            partition.valid = False

    def _Step3(self):

        for partition in self.partitionsL:

            if not partition.valid:
                continue

            if partition.stepNo == 3:
                break

            # Find two columns that my cause split in this step
            p1, p2 = self.DetermineP1P2(partition)

            if p1 == -1 or p2 == -1:
                continue

            try:

                p1Set = set()
                p2Set = set()
                mapRelation1DS = {}
                mapRelation2DS = {}

                # Construct token sets for p1 and p2, dictionary to record the mapping relations between p1 and p2
                for logL in partition.logLL:
                    p1Set.add(logL[p1])
                    p2Set.add(logL[p2])

                    if logL[p1] == logL[p2]:
                        print("Warning: p1 may be equal to p2")

                    if logL[p1] not in mapRelation1DS:
                        mapRelation1DS[logL[p1]] = set()
                    mapRelation1DS[logL[p1]].add(logL[p2])

                    if logL[p2] not in mapRelation2DS:
                        mapRelation2DS[logL[p2]] = set()
                    mapRelation2DS[logL[p2]].add(logL[p1])

                # Construct sets to record the tokens in 1-1, 1-M, M-1 relationships, the left-tokens in p1Set & p2Set
                # are in M-M relationships
                oneToOneS = set()
                oneToMP1D = {}
                oneToMP2D = {}

                # select 1-1 and 1-M relationships
                for p1Token in p1Set:
                    if len(mapRelation1DS[p1Token]) == 1:
                        if len(mapRelation2DS[list(mapRelation1DS[p1Token])[0]]) == 1:
                            oneToOneS.add(p1Token)

                    else:
                        isOneToM = True

                        for p2Token in mapRelation1DS[p1Token]:
                            if len(mapRelation2DS[p2Token]) != 1:
                                isOneToM = False
                                break
                        if isOneToM:
                            oneToMP1D[p1Token] = 0

                # delete the tokens which are picked to 1-1 and 1-M relationships from p1Set, so that the left are M-M
                for deleteToken in oneToOneS:
                    p1Set.remove(deleteToken)
                    p2Set.remove(list(mapRelation1DS[deleteToken])[0])

                for deleteToken in oneToMP1D:
                    for deleteTokenP2 in mapRelation1DS[deleteToken]:
                        p2Set.remove(deleteTokenP2)
                    p1Set.remove(deleteToken)

                # select M-1 relationships
                for p2Token in p2Set:
                    if len(mapRelation2DS[p2Token]) != 1:
                        isOneToM = True
                        for p1Token in mapRelation2DS[p2Token]:
                            if len(mapRelation1DS[p1Token]) != 1:
                                isOneToM = False
                                break
                        if isOneToM:
                            oneToMP2D[p2Token] = 0

                # delete the tokens which are picked to M-1 relationships from p2Set, so that the left are M-M
                for deleteToken in oneToMP2D:
                    p2Set.remove(deleteToken)
                    for deleteTokenP1 in mapRelation2DS[deleteToken]:
                        p1Set.remove(deleteTokenP1)

                # calculate the #Lines_that_match_S
                for logL in partition.logLL:
                    if logL[p1] in oneToMP1D:
                        oneToMP1D[logL[p1]] += 1

                    if logL[p2] in oneToMP2D:
                        oneToMP2D[logL[p2]] += 1

            except KeyError as er:
                print(er)
                print("error: " + str(p1) + "\t" + str(p2))

            newPartitionsD = {}
            if partition.stepNo == 2:
                newPartitionsD["dumpKeyforMMrelationInStep2__"] = Partition(
                    stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                )
            # Split partition
            for logL in partition.logLL:
                # If is 1-1
                if logL[p1] in oneToOneS:
                    if logL[p1] not in newPartitionsD:
                        newPartitionsD[logL[p1]] = Partition(
                            stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                        )
                    newPartitionsD[logL[p1]].logLL.append(logL)
                    newPartitionsD[logL[p1]].numOfLogs += 1

                # This part can be improved. The split_rank can be calculated once.
                # If is 1-M
                elif logL[p1] in oneToMP1D:
                    split_rank = self.Get_Rank_Posistion(
                        len(mapRelation1DS[logL[p1]]), oneToMP1D[logL[p1]], True
                    )
                    if split_rank == 1:
                        if logL[p1] not in newPartitionsD:
                            newPartitionsD[logL[p1]] = Partition(
                                stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                            )
                        newPartitionsD[logL[p1]].logLL.append(logL)
                        newPartitionsD[logL[p1]].numOfLogs += 1
                    else:
                        if logL[p2] not in newPartitionsD:
                            newPartitionsD[logL[p2]] = Partition(
                                stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                            )
                        newPartitionsD[logL[p2]].logLL.append(logL)
                        newPartitionsD[logL[p2]].numOfLogs += 1

                # If is M-1
                elif logL[p2] in oneToMP2D:
                    split_rank = self.Get_Rank_Posistion(
                        len(mapRelation2DS[logL[p2]]), oneToMP2D[logL[p2]], False
                    )
                    if split_rank == 1:
                        if logL[p1] not in newPartitionsD:
                            newPartitionsD[logL[p1]] = Partition(
                                stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                            )
                        newPartitionsD[logL[p1]].logLL.append(logL)
                        newPartitionsD[logL[p1]].numOfLogs += 1
                    else:
                        if logL[p2] not in newPartitionsD:
                            newPartitionsD[logL[p2]] = Partition(
                                stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                            )
                        newPartitionsD[logL[p2]].logLL.append(logL)
                        newPartitionsD[logL[p2]].numOfLogs += 1

                # M-M
                else:
                    if partition.stepNo == 2:
                        newPartitionsD["dumpKeyforMMrelationInStep2__"].logLL.append(
                            logL
                        )
                        newPartitionsD["dumpKeyforMMrelationInStep2__"].numOfLogs += 1
                    else:
                        if len(p1Set) < len(p2Set):
                            if logL[p1] not in newPartitionsD:
                                newPartitionsD[logL[p1]] = Partition(
                                    stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                                )
                            newPartitionsD[logL[p1]].logLL.append(logL)
                            newPartitionsD[logL[p1]].numOfLogs += 1
                        else:
                            if logL[p2] not in newPartitionsD:
                                newPartitionsD[logL[p2]] = Partition(
                                    stepNo=3, numOfLogs=0, lenOfLogs=partition.lenOfLogs
                                )
                            newPartitionsD[logL[p2]].logLL.append(logL)
                            newPartitionsD[logL[p2]].numOfLogs += 1

            if (
                "dumpKeyforMMrelationInStep2__" in newPartitionsD
                and newPartitionsD["dumpKeyforMMrelationInStep2__"].numOfLogs == 0
            ):
                newPartitionsD["dumpKeyforMMrelationInStep2__"].valid = False
            # Add all the new partitions to collection
            for key in newPartitionsD:
                if (
                    self.para.PST != 0
                    and 1.0 * newPartitionsD[key].numOfLogs / partition.numOfLogs
                    < self.para.PST
                ):
                    self.partitionsL[0].logLL += newPartitionsD[key].logLL
                    self.partitionsL[0].numOfLogs += newPartitionsD[key].numOfLogs
                else:
                    self.partitionsL.append(newPartitionsD[key])

            partition.valid = False

    def _Step4(self):
        self.partitionsL[0].valid = False
        if self.para.PST == 0 and self.partitionsL[0].numOfLogs != 0:
            event = Event(["Outlier"])
            event.eventCount = self.partitionsL[0].numOfLogs
            self.eventsL.append(event)

            for logL in self.partitionsL[0].logLL:
                logL.append(str(event.eventId))

        for partition in self.partitionsL:
            if not partition.valid:
                continue

            if partition.numOfLogs == 0:
                print(str(partition.stepNo) + "\t")

            uniqueTokensCountLS = []
            for columnIdx in range(partition.lenOfLogs):
                uniqueTokensCountLS.append(set())

            for logL in partition.logLL:
                for columnIdx in range(partition.lenOfLogs):
                    uniqueTokensCountLS[columnIdx].add(logL[columnIdx])

            e = copy.deepcopy(partition.logLL[0])[: partition.lenOfLogs]

            for columnIdx in range(partition.lenOfLogs):
                if len(uniqueTokensCountLS[columnIdx]) == 1:
                    continue
                else:
                    e[columnIdx] = "<*>"

            event = Event(e)
            event.eventCount = partition.numOfLogs

            self.eventsL.append(event)

            for logL in partition.logLL:
                logL.append(str(event.eventId))

    def _getOutput(self):
        if self.para.PST == 0 and self.partitionsL[0].numOfLogs != 0:
            for logL in self.partitionsL[0].logLL:
                self.output.append(logL[-2:] + logL[:-2])
        for partition in self.partitionsL:
            if not partition.valid:
                continue
            for logL in partition.logLL:
                self.output.append(logL[-2:] + logL[:-2])

    """
    For 1-M and M-1 mappings, you need to decide whether M side are constants or variables. 
    This method is to decide which side to split

    cardOfS           : The number of unique values in this set
    Lines_that_match_S: The number of lines that have these values
    one_m             : If the mapping is 1-M, this value is True. Otherwise, False
    """

    def Get_Rank_Posistion(self, cardOfS, Lines_that_match_S, one_m):
        try:
            distance = 1.0 * cardOfS / Lines_that_match_S
        except ZeroDivisionError as er1:
            print(er1)
            print(
                "cardOfS: "
                + str(cardOfS)
                + "\t"
                + "Lines_that_match_S: "
                + str(Lines_that_match_S)
            )

        if distance <= self.para.lowerBound:
            if one_m:
                split_rank = 2
            else:
                split_rank = 1
        elif distance >= self.para.upperBound:
            if one_m:
                split_rank = 1
            else:
                split_rank = 2
        else:
            if one_m:
                split_rank = 1
            else:
                split_rank = 2

        return split_rank

    def DetermineP1P2(self, partition):
        if partition.lenOfLogs > 2:
            count_1 = 0

            uniqueTokensCountLS = []
            for columnIdx in range(partition.lenOfLogs):
                uniqueTokensCountLS.append(set())

            for logL in partition.logLL:
                for columnIdx in range(partition.lenOfLogs):
                    uniqueTokensCountLS[columnIdx].add(logL[columnIdx])

            # Count how many columns have only one unique term
            for columnIdx in range(partition.lenOfLogs):
                if len(uniqueTokensCountLS[columnIdx]) == 1:
                    count_1 += 1

            # If the columns with unique term more than a threshold, we return (-1, -1) to skip step 3
            GC = 1.0 * count_1 / partition.lenOfLogs

            if GC < self.para.CT:
                return self.Get_Mapping_Position(partition, uniqueTokensCountLS)
            else:
                return (-1, -1)

        elif partition.lenOfLogs == 2:
            return (0, 1)
        else:
            return (-1, -1)

    def Get_Mapping_Position(self, partition, uniqueTokensCountLS):
        p1 = p2 = -1

        # Caculate #unqiueterms in each column, and record how many column with each #uniqueterms
        numOfUniqueTokensD = {}
        for columnIdx in range(partition.lenOfLogs):
            if len(uniqueTokensCountLS[columnIdx]) not in numOfUniqueTokensD:
                numOfUniqueTokensD[len(uniqueTokensCountLS[columnIdx])] = 0
            numOfUniqueTokensD[len(uniqueTokensCountLS[columnIdx])] += 1

        if partition.stepNo == 2:

            # Find the largest card and second largest card
            maxIdx = secondMaxIdx = -1
            maxCount = secondMaxCount = 0
            for key in numOfUniqueTokensD:
                if key == 1:
                    continue
                if numOfUniqueTokensD[key] > maxCount:
                    secondMaxIdx = maxIdx
                    secondMaxCount = maxCount
                    maxIdx = key
                    maxCount = numOfUniqueTokensD[key]
                elif (
                    numOfUniqueTokensD[key] > secondMaxCount
                    and numOfUniqueTokensD[key] != maxCount
                ):
                    secondMaxIdx = key
                    secondMaxCount = numOfUniqueTokensD[key]

            # If the frequency of the freq_card>1 then
            if maxCount > 1:
                for columnIdx in range(partition.lenOfLogs):
                    if len(uniqueTokensCountLS[columnIdx]) == maxIdx:
                        if p1 == -1:
                            p1 = columnIdx
                        else:
                            p2 = columnIdx
                            break

            #                 for columnIdx in range(partition.lenOfLogs):
            #                     if p2 != -1:
            #                         break
            #                     if numOfUniqueTokensD[len(uniqueTokensCountLS[columnIdx])] == secondMaxCount:
            #                         p2 = columnIdx
            #                         break

            # If the frequency of the freq_card==1 then
            else:
                for columnIdx in range(partition.lenOfLogs):
                    if len(uniqueTokensCountLS[columnIdx]) == maxIdx:
                        p1 = columnIdx
                        break

                for columnIdx in range(partition.lenOfLogs):
                    if len(uniqueTokensCountLS[columnIdx]) == secondMaxIdx:
                        p2 = columnIdx
                        break

            if p1 == -1 or p2 == -1:
                return (-1, -1)
            else:
                return (p1, p2)

        # If it is from step 1
        else:
            minIdx = secondMinIdx = -1
            minCount = secondMinCount = sys.maxsize
            for key in numOfUniqueTokensD:
                if numOfUniqueTokensD[key] < minCount:
                    secondMinIdx = minIdx
                    secondMinCount = minCount
                    minIdx = key
                    minCount = numOfUniqueTokensD[key]
                elif (
                    numOfUniqueTokensD[key] < secondMinCount
                    and numOfUniqueTokensD[key] != minCount
                ):
                    secondMinIdx = key
                    secondMinCount = numOfUniqueTokensD[key]

            for columnIdx in range(len(uniqueTokensCountLS)):
                if numOfUniqueTokensD[len(uniqueTokensCountLS[columnIdx])] == minCount:
                    if p1 == -1:
                        p1 = columnIdx
                        break

            for columnIdx in range(len(uniqueTokensCountLS)):
                if (
                    numOfUniqueTokensD[len(uniqueTokensCountLS[columnIdx])]
                    == secondMinCount
                ):
                    p2 = columnIdx
                    break

            return (p1, p2)

    def PrintPartitions(self):
        for idx in range(len(self.partitionsL)):
            print(
                "Partition {}:(from step {})    Valid:{}".format(
                    idx, self.partitionsL[idx].stepNo, self.partitionsL[idx].valid
                )
            )

            for log in self.partitionsL[idx].logLL:
                print(log)

    def PrintEventStats(self):
        for event in self.eventsL:
            if event.eventCount > 1:
                print(str(event.eventId) + "\t" + str(event.eventCount))
                print(event.eventStr)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass
from datetime import datetime

from logai.algorithms.algo_interfaces import AnomalyDetectionAlgo
from logai.config_interfaces import Config
from logai.utils import constants

from merlion.models.anomaly.forecast_based.ets import (
    ETSDetector as MerlionETSDetector,
    ETSDetectorConfig,
)

from logai.utils.functions import pd_to_timeseries
from logai.algorithms.factory import factory


@dataclass
class ETSDetectorParams(Config):
    """
    ETS Anomaly Detector Parameters. For more details of ETS parameters see 
    https://opensource.salesforce.com/Merlion/v1.0.2/merlion.models.forecast.html#module-merlion.models.forecast.ets.

    :param max_forecast_steps: Number of steps we would like to forecast for.
    :param target_seq_index: The index of the univariate (amongst all univariates in a general multivariate
        time series) whose value we would like to forecast.
    :param error: The error term. “add” or “mul”.
    :param trend: The trend component. “add”, “mul” or None.
    :param damped_trend: Whether or not an included trend component is damped.
    :param seasonal: The seasonal component. “add”, “mul” or None.
    :param seasonal_periods: The length of the seasonality cycle. ‘auto’ indicates automatically select the
        seasonality cycle. If no seasonality exists, change seasonal to None.
    :param refit: if ``True``, refit the full ETS model when ``time_series_prev`` is given to the forecast method
        (slower). If ``False``, simply perform exponential smoothing (faster).
    """

    max_forecast_steps: int = None
    target_seq_index: int = None
    error: str = "add"
    trend: str = "add"
    damped_trend: bool = True
    seasonal: str = "add"
    seasonal_periods: str = None
    refit: bool = True
    kwargs: dict = {}


@factory.register("detection", "ets", ETSDetectorParams)
class ETSDetector(AnomalyDetectionAlgo):
    """ETS Anomaly Detector. This is a wrapper for the ETS based Anomaly Detector from Merlion library 
    https://opensource.salesforce.com/Merlion/v1.0.2/merlion.models.forecast.html#module-merlion.models.forecast.ets
    This current version only supports anomaly detection of the constants.LOGLINE_COUNTS (i.e. frequency count of log events)
    """

    def __init__(self, params: ETSDetectorParams):
        ets_config = ETSDetectorConfig(
            max_forecast_steps=params.max_forecast_steps,
            target_seq_index=params.target_seq_index,
            error=params.error,
            trend=params.trend,
            damped_trend=params.damped_trend,
            seasonal=params.seasonal,
            seasonal_periods=params.seasonal_periods,
            refit=params.refit,
            **params.kwargs
        )

        self.model = MerlionETSDetector(ets_config)

    def fit(self, log_features: pd.DataFrame):
        """
        Fit method to train ETS Anomaly Detector.

        :param log_features: A log feature dataframe that must only contain two columns
            ['timestamp': datetime, constants.LOGLINE_COUNTS: int].
        :return: train_scores: The anomaly scores dataframe
            ['index':log_features.index, 'timestamps': datetime, 'anom_score': scores, 'trainval':
            whether it is training set.
        """
        self._is_valid_ts_df(log_features)
        index = log_features.index
        time_series = pd_to_timeseries(log_features)
        train_scores = self.model.train(time_series).to_pd()

        # ETS interpolates missing timestamps, we need to drop those here
        train_scores = train_scores.loc[time_series.to_pd().index]
        train_scores[constants.LOG_TIMESTAMPS] = train_scores.index
        train_scores["trainval"] = True
        train_scores.index = index
        return train_scores

    def predict(self, log_features: pd.DataFrame):
        """
        Predicts anomaly scores for log_feature["timestamp", constants.LOGLINE_COUNTS].

        :param log_features: A log feature dataframe that must only contain two columns
            ['timestamp': datetime, constants.LOGLINE_COUNTS: int].
        :return: test_scores: The anomaly scores dataframe
            ['index':log_features.index, 'timestamps': datetime, 'anom_score': scores, 'trainval':
            whether it is training set.
        """
        self._is_valid_ts_df(log_features)
        index = log_features.index
        time_series = pd_to_timeseries(log_features)
        test_scores = self.model.get_anomaly_label(time_series).to_pd()
        # ETS interpolates missing timestamps, we need to drop those here
        test_scores = test_scores.loc[time_series.to_pd().index]
        test_scores[constants.LOG_TIMESTAMPS] = test_scores.index
        test_scores["trainval"] = False
        test_scores.index = index
        return test_scores

    @staticmethod
    def _is_valid_ts_df(log_feature):
        columns = log_feature.columns.values

        for c in columns:
            if c not in [constants.LOG_TIMESTAMPS, constants.LOG_COUNTS]:
                raise ValueError(
                    "log feature dataframe must only contain two columns ['{}': datetime, '{}': int]".format(
                        constants.LOG_TIMESTAMPS, constants.LOG_COUNTS
                    )
                    + "Current columns: {}".format(columns)
                )

        if constants.LOG_TIMESTAMPS not in columns:
            raise ValueError(
                "dataframe must contain {} column".format(constants.LOG_TIMESTAMPS)
            )

        if constants.LOG_COUNTS not in columns:
            raise ValueError(
                "dataframe must contain {} column".format(constants.LOG_COUNTS)
            )

        for ts in log_feature[constants.LOG_TIMESTAMPS]:
            if not isinstance(ts, datetime):
                raise ValueError("{} must be datetime".format(constants.LOG_TIMESTAMPS))



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from logai.analysis.anomaly_detector import AnomalyDetector, AnomalyDetectionConfig
from logai.applications.application_interfaces import WorkFlowConfig
from logai.dataloader.data_loader import FileDataLoader
from logai.dataloader.data_model import LogRecordObject
from logai.dataloader.openset_data_loader import OpenSetDataLoader
from logai.information_extraction.categorical_encoder import (
    CategoricalEncoder,
    CategoricalEncoderConfig,
)
from logai.information_extraction.feature_extractor import FeatureExtractor
from logai.information_extraction.log_parser import LogParser
from logai.information_extraction.log_vectorizer import LogVectorizer
from logai.preprocess.partitioner import PartitionerConfig, Partitioner
from logai.preprocess.preprocessor import Preprocessor
from logai.utils import constants, evaluate


class LogAnomalyDetection:
    """This is a workflow for log anomaly detection. 
    """
    def __init__(self, config: WorkFlowConfig):
        self.config = config
        self._timestamps = pd.DataFrame()
        self._attributes = pd.DataFrame()
        self._feature_df = pd.DataFrame()
        self._counter_df = pd.DataFrame()
        self._loglines = pd.DataFrame()
        self._log_templates = pd.DataFrame()
        self._ad_results = pd.DataFrame()
        self._labels = pd.DataFrame()
        self._index_group = pd.DataFrame()
        self._loglines_with_anomalies = pd.DataFrame()
        self._group_anomalies = None

        return

    @property
    def timestamps(self):
        return self._timestamps

    @property
    def loglines(self):
        return self._loglines

    @property
    def log_templates(self):
        return self._log_templates

    @property
    def attributes(self):
        return self._attributes

    @property
    def results(self):
        res = (
            self._loglines_with_anomalies.join(self.attributes)
            .join(self.timestamps)
            .join(self.event_group)
        )
        return res

    @property
    def anomaly_results(self):

        return self.results[self.results["is_anomaly"]]

    @property
    def anomaly_labels(self):
        return self._labels

    @anomaly_labels.setter
    def anomaly_labels(self, labels):
        self._labels = labels

    @property
    def event_group(self):
        event_index_map = dict()
        for group_id, indices in self._index_group["event_index"].items():
            for i in indices:
                event_index_map[i] = group_id

        event_index = pd.Series(event_index_map).rename("group_id")
        return event_index

    @property
    def feature_df(self):
        return self._feature_df

    @property
    def counter_df(self):
        return self._counter_df

    def evaluation(self):
        if self.anomaly_labels is None:
            raise TypeError

        labels = self.anomaly_labels.to_numpy()
        pred = np.array([1 if r else 0 for r in self.results["is_anomaly"]])
        return evaluate.get_accuracy_precision_recall(labels, pred)

    def execute(self):
        logrecord = self._load_data()
        # Preprocessor cleans the loglines

        preprocessed_logrecord = self._preprocess(logrecord)

        # Parsing
        loglines = preprocessed_logrecord.body[constants.LOGLINE_NAME]
        parsed_loglines = self._parse(loglines)

        # Feature extraction
        feature_extractor = FeatureExtractor(self.config.feature_extractor_config)

        # Get Counter Set
        self._counter_df = feature_extractor.convert_to_counter_vector(
            timestamps=logrecord.timestamp[constants.LOG_TIMESTAMPS],
            attributes=self.attributes,
        )

        if self.config.anomaly_detection_config.algo_name in constants.COUNTER_AD_ALGO:
            self._counter_df["attribute"] = self._counter_df.drop(
                [constants.LOG_COUNTS, constants.LOG_TIMESTAMPS, constants.EVENT_INDEX],
                axis=1,
            ).apply(lambda x: "-".join(x.astype(str)), axis=1)

            attr_list = self._counter_df["attribute"].unique()
            res = pd.Series()
            for attr in attr_list:
                temp_df = self._counter_df[self._counter_df["attribute"] == attr]
                if temp_df.shape[0] < constants.MIN_TS_LENGTH:
                    anom_score = np.repeat(0.0, temp_df.shape[0])
                    res = res.append(pd.Series(anom_score, index=temp_df.index))
                else:
                    train, test = train_test_split(
                        temp_df[[constants.LOG_TIMESTAMPS, constants.LOG_COUNTS]],
                        shuffle=False,
                        train_size=0.7,
                    )
                    anomaly_detector = AnomalyDetector(
                        self.config.anomaly_detection_config
                    )
                    anom_score_training = pd.Series(
                        np.repeat(0.0, train.shape[0]), index=train.index
                    )
                    anomaly_detector.fit(train)
                    anom_score = anomaly_detector.predict(test)

                    res = res.append(anom_score_training)
                    res = res.append(anom_score["anom_score"])
            self._ad_results = pd.DataFrame(res.rename("result"))
            self._index_group = self._counter_df[[constants.EVENT_INDEX]]

        else:
            # Vectorization
            vectorizor = LogVectorizer(self.config.log_vectorizer_config)
            vectorizor.fit(parsed_loglines)

            # Log vector is a pandas.Series
            log_vectors_w2v = vectorizor.transform(parsed_loglines)

            # Categorical Encoding
            encoder = CategoricalEncoder(self.config.categorical_encoder_config)

            attributes = encoder.fit_transform(logrecord.attributes)
            attributes.columns = logrecord.attributes.columns

            # Get Feature Set
            index_group, feature_df = feature_extractor.convert_to_feature_vector(
                timestamps=logrecord.timestamp[constants.LOG_TIMESTAMPS],
                log_vectors=log_vectors_w2v,
                attributes=attributes,
            )

            self._index_group = index_group[["event_index"]]
            feature_for_anomaly_detection = feature_df.loc[
                :, ~feature_df.columns.isin([constants.LOG_TIMESTAMPS])
            ]
            anomaly_detector = AnomalyDetector(self.config.anomaly_detection_config)

            anomaly_detector.fit(feature_for_anomaly_detection)
            anomalies = anomaly_detector.predict(feature_for_anomaly_detection)[
                "anom_score"
            ]
            self._ad_results = pd.DataFrame(anomalies.rename("result"))

        anomaly_group_indices = self._ad_results[
            self._ad_results["result"] > 0.0
        ].index.values

        anomaly_indices = []

        for indices in self._index_group["event_index"].iloc[anomaly_group_indices]:
            anomaly_indices += indices

        df = pd.DataFrame(self.loglines)
        df["_id"] = df.index.values

        df["is_anomaly"] = [True if i in anomaly_indices else False for i in df["_id"]]
        self._loglines_with_anomalies = df

        return

    def _load_data(self):
        if self.config.open_set_data_loader_config is not None:
            dataloader = OpenSetDataLoader(self.config.open_set_data_loader_config)
            logrecord = dataloader.load_data()
        elif self.config.data_loader_config is not None:
            dataloader = FileDataLoader(self.config.data_loader_config)
            logrecord = dataloader.load_data()
        else:
            raise ValueError(
                "data_loader_config or open_set_data_loader_config is needed to load data."
            )
        return logrecord

    def _preprocess(self, log_record):
        logline = log_record.body[constants.LOGLINE_NAME]

        self._loglines = logline
        self._timestamps = log_record.timestamp
        self._attributes = log_record.attributes.astype(str)

        preprocessor = Preprocessor(self.config.preprocessor_config)
        preprocessed_loglines, _ = preprocessor.clean_log(logline)

        new_log_record = LogRecordObject(
            body=pd.DataFrame(preprocessed_loglines, columns=[constants.LOGLINE_NAME]),
            timestamp=log_record.timestamp,
            attributes=log_record.attributes,
        )

        return new_log_record

    def _parse(self, loglines):

        parser = LogParser(self.config.log_parser_config)
        parsed_results = parser.parse(loglines.dropna())

        parsed_loglines = parsed_results[constants.PARSED_LOGLINE_NAME]

        return parsed_loglines



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass
from sklearn.cluster import KMeans

from logai.algorithms.algo_interfaces import ClusteringAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class KMeansParams(Config):
    """Parameters of the KMeans Clustering algorithm. For more details on the parameters see 
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html.

    :param n_clusters: The number of clusters to form as well as the number of centroids to generate.
    :param init: Method for initialization, i.e., ``{‘k-means++’, ‘random’}``.
    :param n_init: Number of times the k-means algorithm is run with different centroid seeds.
    :param max_iter: Maximum number of iterations of the k-means algorithm for a single run.
    :param tol: Relative tolerance with regards to Frobenius norm of the difference in the cluster
        centers of two consecutive iterations to declare convergence.
    :param verbose: Verbosity mode.
    :param random_state: Determines random number generation for centroid initialization.
    :param copy_x: If copy_x is True (default), then the original data is not modified.
        If False, the original data is modified, and put back before the function returns.
    :param algorithm: K-means algorithm to use, i.e., ``{“lloyd”, “elkan”, “auto”, “full”}``.
    """
    n_clusters: int = 8
    init: str = "k-means++"
    n_init: int = 10
    max_iter: int = 300
    tol: float = 1e-4
    verbose: int = 0
    random_state: int = None
    copy_x: bool = True
    algorithm: str = "auto"


@factory.register("clustering", "kmeans", KMeansParams)
class KMeansAlgo(ClusteringAlgo):
    """
    K-means algorithm for log clustering. This is a wrapper class for K-Means clustering method from
    scikit-learn library https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html.
    """

    def __init__(self, params: KMeansParams):
        self.model = KMeans(
            n_clusters=params.n_clusters,
            init=params.init,
            n_init=params.n_init,
            max_iter=params.max_iter,
            tol=params.tol,
            verbose=params.verbose,
            random_state=params.random_state,
            copy_x=params.copy_x,
            algorithm=params.algorithm,
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Fits a K-means model.

        :param log_features: The log features for training
        """
        self.model.fit(log_features)

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """
        Predicts using trained K-means model.

        :param log_features: The log features for inference.
        :return: A pandas series of cluster labels.
        """
        res = self.model.predict(log_features)
        return pd.Series(res, index=log_features.index)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from transformers import BertForMaskedLM, DataCollatorWithPadding
from datasets import Dataset as HFDataset
import numpy as np
import pandas as pd
from .eval_metric_utils import compute_metrics
from .predict_utils import Predictor, PredictionLabelSmoother
from transformers import TrainingArguments
import logging
import os
from .configs import LogBERTConfig
from logai.utils import constants
from .tokenizer_utils import (
    get_special_tokens,
    get_special_token_ids,
    get_tokenizer,
    get_mask_id,
)


class LogBERTPredict:
    """Class for running inference on logBERT model for unsupervised log anomaly detection.

    :param config: config object describing the parameters of logbert model.
    """

    def __init__(self, config: LogBERTConfig):

        self.config = config

        self.model_dirpath = os.path.join(
            self.config.output_dir, self.config.model_name
        )

        self.model = None
        self.tokenizer = get_tokenizer(self.config.tokenizer_dirpath)
        self.special_tokens = get_special_tokens()
        self.special_token_ids = get_special_token_ids(
            self.tokenizer
        )  # [self.tokenizer.convert_tokens_to_ids(x) for x in special_tokens]
        self.mask_id = get_mask_id(self.tokenizer)

        self.data_collator = DataCollatorWithPadding(
            tokenizer=self.tokenizer,
            padding="max_length",
            max_length=self.config.max_token_len,
        )

        print("initialized data collator")
        self.predictor_args = TrainingArguments(
            self.model_dirpath,
            per_device_eval_batch_size=self.config.per_device_eval_batch_size,
            eval_accumulation_steps=self.config.eval_accumulation_steps,
            resume_from_checkpoint=True,
        )

    def _generate_masked_input(self, examples, indices):
        input_ids = examples["input_ids"][0]
        attention_masks = examples["attention_mask"][0]
        token_type_ids = examples["token_type_ids"][0]
        index = indices[0]
        input_ids = np.array(input_ids)

        sliding_window_diag = np.eye(input_ids.shape[0])

        if sliding_window_diag.shape[0] == 0:
            return examples
        mask = 1 - np.isin(input_ids, self.special_token_ids).astype(np.int64)
        sliding_window_diag = sliding_window_diag * mask
        sliding_window_diag = sliding_window_diag[
            ~np.all(sliding_window_diag == 0, axis=1)
        ]
        num_sections = int(sliding_window_diag.shape[0] / self.config.mask_ngram)
        if num_sections <= 0:
            num_sections = sliding_window_diag.shape[0]
        sliding_window_diag = np.array_split(sliding_window_diag, num_sections, axis=0)
        diag = np.array([np.sum(di, axis=0) for di in sliding_window_diag])

        input_rpt = np.tile(input_ids, (diag.shape[0], 1))
        labels = np.copy(input_rpt)
        input_ids_masked = (input_rpt * (1 - diag) + diag * self.mask_id).astype(
            np.int64
        )
        attention_masks = np.tile(
            np.array(attention_masks), (input_ids_masked.shape[0], 1)
        )
        token_type_ids = np.tile(
            np.array(token_type_ids), (input_ids_masked.shape[0], 1)
        )
        labels[
            input_ids_masked != self.mask_id
        ] = -100  # Need masked LM loss only for tokens with mask_id
        examples = {}
        examples["input_ids"] = input_ids_masked
        examples["attention_mask"] = attention_masks
        examples["token_type_ids"] = token_type_ids
        examples["labels"] = labels
        examples["indices"] = np.array([index] * input_ids_masked.shape[0]).astype(
            np.int64
        )
        return examples

    def load_model(self):
        """Loading logbert model from the model dir path as specified in the logBERTConfig config"""
        checkpoint_dir = "checkpoint-" + str(
            max(
                [
                    int(x.split("-")[1])
                    for x in os.listdir(self.model_dirpath)
                    if x.startswith("checkpoint")
                ]
            )
        )
        model_checkpoint = os.path.abspath(
            os.path.join(self.model_dirpath, checkpoint_dir)
        )
        logging.info("Loading model from {}".format(model_checkpoint))
        self.model = BertForMaskedLM.from_pretrained(model_checkpoint)
        self.model.tokenizer = self.tokenizer  # self.vectorizer.tokenizer
        self.predictor = Predictor(
            model=self.model,
            args=self.predictor_args,
            train_dataset=None,
            eval_dataset=None,
            data_collator=self.data_collator,
        )
        self.predictor.label_smoother = PredictionLabelSmoother(
            epsilon=self.predictor_args.label_smoothing_factor
        )

    def predict(self, test_dataset: HFDataset):
        """Method for running inference on logbert to predict anomalous loglines in test dataset.

        :param test_dataset: test dataset of type huggingface Dataset object.
        :return: dict containing instance-wise loss and scores.
        """
        if not self.model:
            self.load_model()

        test_labels = {k: v for k, v in enumerate(list(test_dataset[constants.LABELS]))}

        if constants.LOG_COUNTS in test_dataset:
            test_counts = {
                k: v for k, v in enumerate(list(test_dataset[constants.LOG_COUNTS]))
            }
        else:
            test_counts = None

        mlm_dataset_test = test_dataset.map(
            self._generate_masked_input,
            with_indices=True,
            batched=True,
            batch_size=1,
            num_proc=1,
            remove_columns=test_dataset.column_names,
        )

        num_shards = self.config.num_eval_shards
        for i in range(num_shards):
            test_masked_lm_shard = mlm_dataset_test.shard(
                num_shards=num_shards, index=i
            )
            test_results = self.predictor.predict(test_masked_lm_shard)
            logging.info(
                "test_loss: {} test_runtime: {} test_samples/s: {}".format(
                    test_results.metrics["test_loss"],
                    test_results.metrics["test_runtime"],
                    test_results.metrics["test_samples_per_second"],
                )
            )

            data_columns = [
                "indices",
                "max_loss",
                "sum_loss",
                "num_loss",
                "top6_loss",
                "top6_max_prob",
                "top6_min_logprob",
                "top6_max_entropy",
            ]
            eval_metrics_per_instance_series = pd.DataFrame(
                np.transpose(
                    np.array(
                        self.predictor.label_smoother.eval_metrics_per_instance,
                        dtype=object,
                    )
                ),
                index=range(
                    len(self.predictor.label_smoother.eval_metrics_per_instance[0])
                ),
                columns=data_columns,
            )
            logging.info(
                "number of original test instances {}".format(
                    len(eval_metrics_per_instance_series.groupby("indices"))
                )
            )

            if i % 2 == 0 and test_labels is not None:
                compute_metrics(
                    eval_metrics_per_instance_series, test_labels, test_counts
                )

        return eval_metrics_per_instance_series



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import os
import pathlib
import json
import pandas as pd
import yaml
import pickle as pkl


def file_exists(path: str):
    """util function to check if file exists
    
    :param path: (str): path to file
    :return: bool: if file exists or not
    """
    return os.path.exists(path)


def read_file(filepath: str):
    """Reads yaml, json, csv or pickle files.

    :param filepath: (str): path to file
    :return: data object containing file contents
    """
    file_type = pathlib.Path(filepath).suffix
    if file_type == ".yaml":
        with open(filepath, "r") as stream:
            try:
                data = yaml.full_load(stream)
            except yaml.YAMLError as exc:
                raise exc
    elif file_type == ".json":
        with open(filepath, "r") as f:
            data = json.load(f)
    elif file_type == ".csv":
        with open(filepath, "r"):
            data = pd.read_csv(filepath)
    elif file_type == ".pkl" or file_type == ".pickle":
        data = pkl.load(open(filepath, "rb"))
    else:
        raise Exception("cannot read filetype", file_type)
    return data



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from logai.algorithms.algo_interfaces import CategoricalEncodingAlgo
from logai.config_interfaces import Config


class OneHotEncodingParams(Config):
    """
    Configuration for One-Hot Encoding. For more details on the parameters see
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html.

    :param categories: Categories (unique values) per feature.
    :param drop: Specifies a methodology to use to drop one of the categories per feature.
    :param dtype: Desired dtype of output.
    :param handle_unknown: Specifies the way unknown categories are handled during transform.
    """
    categories: str = "auto"
    drop: object = None
    dtype: np.float64 = np.float64
    handle_unknown: str = "error"


class OneHotEncoding(CategoricalEncodingAlgo):
    """This is a wrapper class for OneHotEncoder from scikit-learn library
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html.
    """
    def __init__(self, params: OneHotEncodingParams):
        self.model = OneHotEncoder(
            categories=params.categories,
            drop=params.drop,
            sparse=False,  # Default return non-sparse matrix, need to think about how to handle sparse matrix
            dtype=params.dtype,
            handle_unknown=params.handle_unknown,
        )

    def fit_transform(self, log_attributes: pd.DataFrame) -> pd.DataFrame:
        """
        Fits and transforms log attributes into one-hot encoding categories.
        
        :param log_attributes: A list of log attributes in text form.
        :return: A pandas dataframe of categories in on-hot encoding.
        """
        col_names = log_attributes.columns
        if len(col_names) == 1:
            res_col_name_prefix = col_names[0]
        else:
            res_col_name_prefix = "-".join(col_names)
        self.model.fit(log_attributes)
        res = pd.DataFrame(
            self.model.transform(log_attributes), index=log_attributes.index
        )
        res_col_names = ["{}-{}".format(res_col_name_prefix, c) for c in res.columns]
        res.columns = res_col_names
        return res



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import json
import os.path

from attr import dataclass

from logai.config_interfaces import Config
from logai.dataloader.data_loader import FileDataLoader, DataLoaderConfig


def get_config(dataset_name, filepath) -> DataLoaderConfig:
    """
    Retrieves the configuration of open log datasets to load data.
    
    :param dataset_name: The supported log dataset name from ("hdfs", "bgl", "HealthApp").
    :param filepath: The log file path.
    :return: The configuration to load open log datasets.
    """
    config_path = os.path.join(
        os.path.dirname(__file__),
        "openset_configs",
        "{}.json".format(dataset_name.lower()),
    )
    with open(config_path, "r") as f:
        config = DataLoaderConfig.from_dict(json.load(f))
    config.filepath = filepath
    return config


@dataclass
class OpenSetDataLoaderConfig(Config):
    dataset_name: str = None
    filepath: str = None


class OpenSetDataLoader(FileDataLoader):
    def __init__(self, config: OpenSetDataLoaderConfig):
        self._dl_config = get_config(config.dataset_name, config.filepath)
        super().__init__(self._dl_config)
        return

    def load_data(self):
        return super().load_data()

    @property
    def dl_config(self):
        return self._dl_config



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from attr import dataclass

from logai.config_interfaces import Config
from logai.dataloader.data_model import LogRecordObject


@dataclass
class PreprocessorConfig(Config):
    """Config class for Preprocessor.

    :param custom_delimiters_regex: A dictionary of delimiter regex patterns in raw log data.
    :param custom_replace_list: A list of tuples of custom replace patterns in raw log data.
        Each Tuple should be of form ('regex-pattern-to-replace', 'replaced-pattern').
    """
    custom_delimiters_regex: dict = None
    custom_replace_list: list = None


class Preprocessor:
    """
    Preprocess class that contains common preprocess methods.
    """

    def __init__(self, config: PreprocessorConfig):
        self.config = config

    def clean_log(self, loglines: pd.Series) -> pd.Series:
        """Cleans the input log data.

        :param loglines: The raw loglines data to be cleaned .
        :return:pd.Series: The cleaned loglines data .
        """
        cleaned_log = loglines
        terms = pd.DataFrame()
        if self.config.custom_delimiters_regex:
            for reg in self.config.custom_delimiters_regex:
                try:
                    cleaned_log = cleaned_log.replace(
                        to_replace=reg, value=" ", regex=True
                    )
                except:
                    raise RuntimeError(
                        "Cannot replace custom regex delimiter {}".format(reg)
                    )

        if self.config.custom_replace_list:
            for pair in self.config.custom_replace_list:
                # TODO: refactor to tuple or map.
                try:
                    pattern = pair[0]
                    replacement = pair[1]
                    terms[replacement] = cleaned_log.str.findall(pat=pattern)
                    cleaned_log = cleaned_log.replace(
                        to_replace=pattern, value=replacement, regex=True
                    )
                except:
                    raise RuntimeError(
                        "Cannot replace custom regex: {} values: {}".format(
                            pair[0], pair[1]
                        )
                    )
        return cleaned_log, terms

    def group_log_index(self, attributes: pd.DataFrame, by: np.array) -> pd.DataFrame:
        """Groups log attributes (DataFrame) by a list of its fields.

        :param attributes: The log attribute data to be grouped.
        :param by: A list of fields of the log attribute DataFrame object to group by.
        :return: The log attribute data after grouping.
        """
        attributes["group_index"] = attributes.index
        group_index_list = (
            attributes.groupby(by=by).group_index.apply(np.array).reset_index()
        )

        return group_index_list

    def identify_timestamps(self, logrecord: LogRecordObject):
        pass



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from schema import Or, Schema, Optional

config_schema = Schema(
    {
        "dataset_name": str,
        Optional("label_filepath"): str,
        "parse_logline": bool,
        "output_dir": str,
        "output_file_type": str,
        "training_type": str,
        "deduplicate_test": bool,
        Optional("test_data_frac_pos"): float,
        Optional("test_data_frac_neg"): float,
        Optional("train_test_shuffle"): bool,
        "data_loader_config": {
            "filepath": str,
            "reader_args": {"log_format": str},
            "log_type": str,
            "dimensions": {
                "body": list,
                "timestamp": list,
                Optional("labels"): Or(list, None),
                Optional("span_id"): list,
            },
            "datetime_format": str,
            "infer_datetime": bool,
        },
        "preprocessor_config": {
            "custom_delimiters_regex": list,
            "custom_replace_list": list,
        },
        "open_set_partitioner_config": {
            "session_window": bool,
            "sliding_window": int,
            "logsequence_delim": str,
        },
        "log_parser_config": {
            "parsing_algorithm": str,
            "parsing_algo_params": {"sim_th": float, "depth": int},
        },
        "log_vectorizer_config": {
            "algo_name": str,
            "algo_param": {
                Optional("model_name"): str,
                Optional("feature_type"): str,
                Optional("sep_token"): str,
                Optional("max_token_len"): int,
                Optional("embedding_dim"): int,
                Optional("label_type"): str,
                Optional("custom_tokens"): list,
                Optional("output_dir"): str,
                Optional("tokenizer_dirname"): str,
                Optional("vectorizer_model_dirpath"): str,
                Optional("vectorizer_metadata_filepath"): str,
            },
        },
        "nn_anomaly_detection_config": {
            "algo_name": str,
            "algo_params": {
                "model_name": str,
                Optional("learning_rate"): float,
                Optional("embedding_dim"): int,
                Optional("max_token_len"): int,
                Optional("feature_type"): str,
                Optional("label_type"): str,
                Optional("eval_type"): str,
                Optional("batch_size"): int,
                Optional("per_device_train_batch_size"): int,
                Optional("num_train_epochs"): int,
                Optional("save_steps"): int,
                Optional("mask_ngram"): int,
                Optional("output_dir"): str,
                Optional("tokenizer_dirpath"): str,
                Optional("metadata_filepath"): str,
            },
        },
    }
)
